#include "pqOptAirfoil.h"
#include "pqPropertiesPanel.h"
#include "pvOptFuncs.cxx"
#include "pvOptFuncs.h"

#include <QCheckBox>
#include <QGridLayout>
#include <QLabel>
#include <QPushButton>
#include <QString>
#include <QtConcurrent/QtConcurrent>
#include <QFuture>
#include <QByteArray>
#include <QFileDialog>
#include <QMessageBox>

#include <iostream>
#include <cstdlib>
#include <cstring>
#include <sstream>

#include <vtksys/SystemTools.hxx>

#ifdef _WIN32
  #include <windows.h>
#else
  #include <unistd.h>
#endif
      
//-----------------------------------------------------------------------------
pqOptAirfoil::pqOptAirfoil(
  vtkSMProxy* smproxy, vtkSMProperty* smproperty, QWidget* parentObject)
  : Superclass(smproxy, parentObject)
{

  this->setShowLabel(false);

  QGridLayout* gridLayout = new QGridLayout(this);
  gridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  gridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  gridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  gridLayout->setColumnStretch(1, 1);
  QLabel* spacer = new QLabel(this);

  //link case file button
  QPushButton* caseFileHandlerButton = new QPushButton(this);
  caseFileHandlerButton->setText(tr("Select Optimization Folder"));
  connect(caseFileHandlerButton, SIGNAL(clicked()), SLOT(caseFileHandler()));
  gridLayout->addWidget(caseFileHandlerButton);

  //About button
  QPushButton* aboutButton = new QPushButton(this);
  aboutButton->setText(tr("About pvOptAirfoil"));
  connect(aboutButton, SIGNAL(clicked()), SLOT(aboutMessage()));
  gridLayout->addWidget(aboutButton);

  // button to select upper/suction surface
  upperSurfacePath = new QString();
  QPushButton* upperSurfaceSelectButton = new QPushButton(this);
  upperSurfaceSelectButton->setText(tr("Select Upper Surface Profile"));
  connect(upperSurfaceSelectButton, SIGNAL(clicked()), SLOT(selectUpperSurface()));
  gridLayout->addWidget(upperSurfaceSelectButton);

  // button to select lower/pressure surface
  lowerSurfacePath = new QString();
  QPushButton* lowerSurfaceSelectButton = new QPushButton(this);
  lowerSurfaceSelectButton->setText(tr("Select Lower Surface Profile"));
  connect(lowerSurfaceSelectButton, SIGNAL(clicked()), SLOT(selectLowerSurface()));
  gridLayout->addWidget(lowerSurfaceSelectButton);

  // number of mesh cells
  QLabel* numberOfTotalCells = new QLabel(tr("Number of Total Mesh Cells"), this);
  gridLayout->addWidget(numberOfTotalCells);
  nTotalCells = new QLineEdit(this);
  nTotalCells->setText("4000");
  connect(nTotalCells, SIGNAL(editingFinished()), SLOT(calculateAdvancedMeshSettings()));
  gridLayout->addWidget(nTotalCells);

  // first layer mesh length
  QLabel* firstMeshLayerLength = new QLabel(tr("Length of First Mesh Layer"), this);
  gridLayout->addWidget(firstMeshLayerLength);
  yWall = new QLineEdit(this);
  yWall->setText("4e-3");
  gridLayout->addWidget(yWall);

  // number of points to extrude
  QLabel* numberOfExtrusionPoints = new QLabel(tr("Number of Points to Extrude"), this);
  gridLayout->addWidget(numberOfExtrusionPoints);
  NpExtrude = new QLineEdit(this);
  NpExtrude->setText("32");
  connect(NpExtrude, SIGNAL(editingFinished()), SLOT(calculateAdvancedMeshSettings()));
  gridLayout->addWidget(NpExtrude);

  // button to toggle advanced mesh settings
  advancedSettingsEnabled = new bool();
  *advancedSettingsEnabled = false;
  QPushButton* advancedSettingsButton = new QPushButton(tr("Toggle Advanced Mesh Settings"),this);
  connect(advancedSettingsButton, SIGNAL(clicked()), SLOT(toggleAdvancedSettings()));
  gridLayout->addWidget(advancedSettingsButton);
  gridLayout->addWidget(spacer);

  // dX1PS
  QLabel* dX1PS = new QLabel(tr("Lower: First dx from LE"), this);
  gridLayout->addWidget(dX1PS);
  dX1PSValue = new QLineEdit(this);
  dX1PSValue->setText("0.005");
  connect(dX1PSValue, SIGNAL(editingFinished()), SLOT(calculateTotalCells()));
  gridLayout->addWidget(dX1PSValue);

  // Alpha1PS

  // dX2PS
  QLabel* dX2PS = new QLabel(tr("Lower: First dx from TE"), this);
  gridLayout->addWidget(dX2PS);
  dX2PSValue = new QLineEdit(this);
  dX2PSValue->setText("0.002");
  connect(dX2PSValue, SIGNAL(editingFinished()), SLOT(calculateTotalCells()));
  gridLayout->addWidget(dX2PSValue);

  // Alpha2PS

  // dXMaxPS
  QLabel* dXMaxPS = new QLabel(tr("Max dx for Lower"), this);
  gridLayout->addWidget(dXMaxPS);
  dXMaxPSValue = new QLineEdit(this);
  dXMaxPSValue->setText("0.02");
  connect(dXMaxPSValue, SIGNAL(editingFinished()), SLOT(calculateTotalCells()));
  gridLayout->addWidget(dXMaxPSValue);

  // dX1SS
  QLabel* dX1SS = new QLabel(tr("Upper: First dx from LE"), this);
  gridLayout->addWidget(dX1SS);
  dX1SSValue = new QLineEdit(this);
  dX1SSValue->setText("0.005");
  connect(dX1SSValue, SIGNAL(editingFinished()), SLOT(calculateTotalCells()));
  gridLayout->addWidget(dX1SSValue);

  // Alpha1SS

  // dX2SS
  QLabel* dX2SS = new QLabel(tr("Upper: First dx from TE"), this);
  gridLayout->addWidget(dX2SS);
  dX2SSValue = new QLineEdit(this);
  dX2SSValue->setText("0.002");
  connect(dX2SSValue, SIGNAL(editingFinished()), SLOT(calculateTotalCells()));
  gridLayout->addWidget(dX2SSValue);

  // Alpha2SS

  // dXMaxSS
  QLabel* dXMaxSS = new QLabel(tr("Max dx for Upper"), this);
  gridLayout->addWidget(dXMaxSS);
  dXMaxSSValue = new QLineEdit(this);
  dXMaxSSValue->setText("0.02");
  connect(dXMaxSSValue, SIGNAL(editingFinished()), SLOT(calculateTotalCells()));
  gridLayout->addWidget(dXMaxSSValue);

  // NpTE
  QLabel* NpTE = new QLabel(tr("Number of Points in Blunt TE"), this);
  gridLayout->addWidget(NpTE);
  NpTEValue = new QLineEdit(this);
  NpTEValue->setText("5");
  gridLayout->addWidget(NpTEValue);

  // marching distance
  QLabel* marchingDistance = new QLabel(tr("Marching Distance for Extrusion"));
  gridLayout->addWidget(marchingDistance);
  marchingDistanceValue = new QLineEdit(this);
  marchingDistanceValue->setText("20.0");
  gridLayout->addWidget(marchingDistanceValue);

  // button to preprocess
  QPushButton* preProcessingButton = new QPushButton(this);
  preProcessingButton->setText(tr("Generate Mesh"));
  connect(preProcessingButton, SIGNAL(clicked()), SLOT(genMesh()));
  gridLayout->addWidget(preProcessingButton);
  gridLayout->addWidget(spacer);

  // solver
  QLabel* solver = new QLabel(tr("Solver"), this);
  gridLayout->addWidget(solver);
  solverBox = new QComboBox(this);
  solverBox->addItem(tr("DASimpleFoam (incompressible)"));
  solverBox->addItem(tr("DARhoSimpleFoam (subsonic)"));
  solverBox->addItem(tr("DARhoSimpleCFoam (transonic)"));
  connect(solverBox, SIGNAL(currentIndexChanged(int)), SLOT(caseSpecificParameters()));
  gridLayout->addWidget(solverBox);

  // type of run
  QLabel* runTypeToDo = new QLabel(tr("Type of Run"), this);
  gridLayout->addWidget(runTypeToDo);
  runType = new QComboBox(this);
  runType->addItem(tr("Optimize"));
  runType->addItem(tr("Solve Flow"));
  connect(runType, SIGNAL(currentIndexChanged(int)), SLOT(runTypeChanged()));
  gridLayout->addWidget(runType);
  
  // optimizer
  QLabel* optimizer = new QLabel(tr("Optimizer"), this);
  gridLayout->addWidget(optimizer);
  optimizerBox = new QComboBox(this);
  optimizerBox->addItem(tr("ipopt"));
  optimizerBox->addItem(tr("slsqp"));
  optimizerBox->addItem(tr("snopt"));
  gridLayout->addWidget(optimizerBox);

  // mach number
  QLabel* machNumber = new QLabel(tr("Mach Number"), this);
  gridLayout->addWidget(machNumber);
  machNumberValue = new QLineEdit(this);
  connect(machNumberValue, SIGNAL(editingFinished()), SLOT(machWarning()));
  gridLayout->addWidget(machNumberValue);

  // Reynolds number
  QLabel* reynoldsNumber = new QLabel(tr("Reynolds Number"), this);
  gridLayout->addWidget(reynoldsNumber);
  reynoldsNumberValue = new QLineEdit(this);
  connect(reynoldsNumberValue, SIGNAL(editingFinished()), SLOT(scaleMesh()));
  gridLayout->addWidget(reynoldsNumberValue);

  // far field pressure
  QLabel* farFieldPressure = new QLabel(tr("Far Field Pressure"), this);
  gridLayout->addWidget(farFieldPressure);
  farFieldPressureValue = new QLineEdit(this);
  gridLayout->addWidget(farFieldPressureValue);

  // normalization density
  QLabel* normalizationDensity = new QLabel(tr("Normalization Density"), this);
  gridLayout->addWidget(normalizationDensity);
  normalizationDensityValue = new QLineEdit(this);
  gridLayout->addWidget(normalizationDensityValue);
  normalizationDensityValue->setText("1.0");

  // temperature
  QLabel* temperature = new QLabel(tr("Temperature"), this);
  gridLayout->addWidget(temperature);
  temperatureValue = new QLineEdit(this);
  gridLayout->addWidget(temperatureValue);
  temperatureValue->setText("273.15");

  // dynamic viscosity
  QLabel* viscosity = new QLabel(tr("Dynamic Viscosity"), this);
  gridLayout->addWidget(viscosity);
  dynamicViscosityValue = new QLineEdit(this);
  gridLayout->addWidget(dynamicViscosityValue);
  dynamicViscosityValue->setText("1.5e-5");
  connect(dynamicViscosityValue, SIGNAL(editingFinished()), SLOT(scaleReynolds()));

  // dynamic viscosity
  alphaZeroValue = new double;
  *alphaZeroValue = 3.0;
  QLabel* alpha = new QLabel(tr("Angle of Attack"), this);
  gridLayout->addWidget(alpha);
  alphaValue = new QLineEdit(this);
  gridLayout->addWidget(alphaValue);
  alphaValue->setText(std::to_string(*alphaZeroValue).c_str());
  alphaValue->setEnabled(false);

  // turbulence model
  QLabel* turbulenceModel = new QLabel(tr("Turbulence Model"), this);
  gridLayout->addWidget(turbulenceModel);
  turbulenceModelBox = new QComboBox(this);
  turbulenceModelBox->addItem(tr("Spalart Allmaras"));
  turbulenceModelBox->addItem(tr("k-omega SST"));
  turbulenceModelBox->addItem(tr("k epsilon"));
  connect(turbulenceModelBox, SIGNAL(currentIndexChanged(int)), SLOT(caseSpecificParameters()));
  gridLayout->addWidget(turbulenceModelBox);

  // button to toggle advanced settings
  advancedTurbulenceEnabled = new bool();
  *advancedTurbulenceEnabled = false;
  QPushButton* advancedTurbulenceButton = new QPushButton(tr("Toggle Advanced Turbulence Settings"),this);
  connect(advancedTurbulenceButton, SIGNAL(clicked()), SLOT(toggleAdvancedTurbulence()));
  gridLayout->addWidget(advancedTurbulenceButton);
  gridLayout->addWidget(spacer);

  // eddy viscosity ratio
  QLabel* viscosityRatio = new QLabel(tr("Eddy Viscosity Ratio"), this);
  gridLayout->addWidget(viscosityRatio);
  eddyViscosityRatioValue = new QLineEdit(this);
  gridLayout->addWidget(eddyViscosityRatioValue);
  eddyViscosityRatioValue->setText("10.0");

  // turbulent intensity i
  QLabel* turbulentIntensity = new QLabel(tr("Turbulent Intensity"), this);
  gridLayout->addWidget(turbulentIntensity);
  turbulentIntensityValue = new QLineEdit(this);
  gridLayout->addWidget(turbulentIntensityValue);
  turbulentIntensityValue->setText("0.01");

  // target lift coefficient
  QLabel* targetLiftCoefficient = new QLabel(tr("Target Lift Coefficient"), this);
  gridLayout->addWidget(targetLiftCoefficient);
  targetLiftCoefficientValue = new QLineEdit(this);
  targetLiftCoefficientValue->setText("0.5");
  gridLayout->addWidget(targetLiftCoefficientValue);

  // number of FFD points
  QLabel* numberFFDPoints = new QLabel(tr("Number of FFD Points"), this);
  gridLayout->addWidget(numberFFDPoints);
  numberFFDPointsValue = new QLineEdit(this);
  gridLayout->addWidget(numberFFDPointsValue);
  numberFFDPointsValue->setText("20");

  // maximum major iterations
  QLabel* maximumIterations = new QLabel(tr("Maximum Optimization Iterations"), this);
  gridLayout->addWidget(maximumIterations);
  maximumIterationsValue = new QLineEdit(this);
  gridLayout->addWidget(maximumIterationsValue);
  maximumIterationsValue->setText("50");

  // Flow Iterations
  QLabel* flowIterations = new QLabel(tr("Maximum Flow Iterations"),this);
  gridLayout->addWidget(flowIterations);
  flowIterationsValue = new QLineEdit(this);
  gridLayout->addWidget(flowIterationsValue);
  flowIterationsValue->setText("2000");

  // button to toggle advanced parameters
  advancedRunscriptEnabled = new bool();
  *advancedRunscriptEnabled = false;
  QPushButton* advancedRunscriptButton = new QPushButton(tr("Toggle Advanced Runscript Settings"),this);
  connect(advancedRunscriptButton, SIGNAL(clicked()), SLOT(toggleAdvancedRunscript()));
  gridLayout->addWidget(advancedRunscriptButton);
  gridLayout->addWidget(spacer);

  // flow convergence tolerance
  QLabel* flowConvergenceTolerance = new QLabel(tr("Flow Convergence Tolerance"),this);
  gridLayout->addWidget(flowConvergenceTolerance);
  flowConvergenceToleranceValue = new QLineEdit(this);
  gridLayout->addWidget(flowConvergenceToleranceValue);
  flowConvergenceToleranceValue->setText("1.0e-8");

  // adjoint convergence Tolerance
  QLabel* adjointConvergenceTolerance = new QLabel(tr("Adjoint Convergence Tolerance"), this);
  gridLayout->addWidget(adjointConvergenceTolerance);
  adjointConvergenceToleranceValue = new QLineEdit(this);
  gridLayout->addWidget(adjointConvergenceToleranceValue);
  adjointConvergenceToleranceValue->setText("1.0e-6");

  //feasibility Tolerance
  QLabel* feasibilityTol = new QLabel(tr("Feasibility Tolerance"), this);
  gridLayout->addWidget(feasibilityTol);
  feasibilityTolerance = new QLineEdit(this);
  gridLayout->addWidget(feasibilityTolerance);
  feasibilityTolerance->setText("1.0e-6");

  //Optimality Tolerance
  QLabel* optimalityTol = new QLabel(tr("Optimality Tolerance"), this);
  gridLayout->addWidget(optimalityTol);
  optimalityTolerance = new QLineEdit(this);
  gridLayout->addWidget(optimalityTolerance);
  optimalityTolerance->setText("1.0e-6");

  // minumum thickness constraint
  QLabel* minimumThickness = new QLabel(tr("Minimum Relative Thickness"), this);
  gridLayout->addWidget(minimumThickness);
  minimumThicknessValue = new QLineEdit(this);
  gridLayout->addWidget(minimumThicknessValue);
  minimumThicknessValue->setText("0.8");

  // maximum thickness constraint
  QLabel* maximumThickness = new QLabel(tr("Maximum Relative Thickness"), this);
  gridLayout->addWidget(maximumThickness);
  maximumThicknessValue = new QLineEdit(this);
  gridLayout->addWidget(maximumThicknessValue);
  maximumThicknessValue->setText("3.0");

  // minimum volume constraint
  QLabel* minimumVolume = new QLabel(tr("Minimum Relative Volume"), this);
  gridLayout->addWidget(minimumVolume);
  minimumVolumeValue = new QLineEdit(this);
  gridLayout->addWidget(minimumVolumeValue);
  minimumVolumeValue->setText("1.0");

  // maximum volume constraint
  QLabel* maximumVolume = new QLabel(tr("Maximum Relative Volume"), this);
  gridLayout->addWidget(maximumVolume);
  maximumVolumeValue = new QLineEdit(this);
  gridLayout->addWidget(maximumVolumeValue);
  maximumVolumeValue->setText("3.0");

  // custom daOptions
  QLabel* customDAConfig = new QLabel(tr("Custom DaOptions - Comma Seperated"));
  gridLayout->addWidget(customDAConfig);
  customDaConfigValue = new QLineEdit(this);
  gridLayout->addWidget(customDaConfigValue);
  customDaConfigValue->setText("");

  // button to write runscript.py
  QPushButton* writeRunScriptButton = new QPushButton(this);
  writeRunScriptButton->setText(tr("Write Run Script"));
  connect(writeRunScriptButton, SIGNAL(clicked()), SLOT(writeRunScript()));
  gridLayout->addWidget(writeRunScriptButton);

  // post-processing button
  QPushButton* postProcessButton = new QPushButton(this);
  postProcessButton->setText(tr("View Optimization Data"));
  connect(postProcessButton, SIGNAL(clicked()), SLOT(postProcessHandler()));
  gridLayout->addWidget(postProcessButton);

  // since there's no such thing a "editing" and 'editing done' for checkboxes.
  this->setChangeAvailableAsChangeFinished(true);
  
  goodDocker = new bool();
  preProcessingComplete = new bool();
  meshGenComplete = new bool();
  processingComplete = new bool();
  profileReferenced = new bool();
  caseFileDir = new QString();
  chordValue = new double();
  originalChordValue = new double();
  scaleFactor = new double();
  optOutputFile = new std::string();
  pyGuiPath = new std::string();
  alphaTitle = new std::string();
  optTitle = new std::string();
  simTitle = new std::string();
  daVersion = new std::string();
  printIntervalValue = new int();
  referenceAreaValue = new double();
  optimizationRunning = new bool();

  *goodDocker = false;
  *preProcessingComplete = false;
  *meshGenComplete = false;
  *processingComplete = false;
  *profileReferenced = false;
  *originalChordValue = 1.0;
  *chordValue = 1.0;
  *optOutputFile = "logOpt.txt";
  *daVersion = "2.2.9";

  calculateAdvancedMeshSettings();
  caseSpecificParameters();
  advancedRunscript();

  prtValue = new double();  
  *prtValue = 1.0;
  *optimizationRunning = false;
  *printIntervalValue = 10;

  preProcessingFile = "/preProcessing.sh";
  logMeshGenOut = "logMeshGeneration.txt";
}

//-----------------------------------------------------------------------------
pqOptAirfoil::~pqOptAirfoil()
{
}

// toggle advanced settings
void pqOptAirfoil::toggleAdvancedSettings()
{
  if (*advancedSettingsEnabled)
  {
    *advancedSettingsEnabled = false;
  }
  else
  {
    *advancedSettingsEnabled = true;
  }
  calculateAdvancedMeshSettings();
}

void pqOptAirfoil::runTypeChanged()
{
  bool enabled = (runType->currentText()==tr("Optimize") ? true : false);
  if (enabled) { alphaValue->setText(std::to_string(*alphaZeroValue).c_str()); }
  optimizerBox->setEnabled(enabled);
  targetLiftCoefficientValue->setEnabled(enabled);
  numberFFDPointsValue->setEnabled(enabled);
  maximumIterationsValue->setEnabled(enabled);
  alphaValue->setEnabled(!enabled);
}

// check if docker is good
void pqOptAirfoil::dockerMessage()
{
  dockerFound info = pvOptFuncs::checkDocker(rootPath);
  *goodDocker = info.goodDocker;

  QMessageBox dockerInfo;
  dockerInfo.setText(info.success ? (*goodDocker ? tr("       Docker Successfully Found       ") : tr("      Docker Not Found     ")) : tr("       Error -- No Docker Info       "));
  dockerInfo.setStandardButtons(QMessageBox::Ok);
  dockerInfo.setInformativeText(*goodDocker ? tr(info.infoStr.c_str()) : tr("Without Docker, local mesh generation/transformation and running optimization is not possible"));
  int ret = dockerInfo.exec();
}

void pqOptAirfoil::aboutMessage()
{
  QMessageBox about;
  about.setStyleSheet("QLabel{width: 400px; min-width: 600px; max-width: 400px;}");
  about.setText("<b><p align='center'><big>pvOptAirfoil<br></b></p></big>");
  about.setStandardButtons(QMessageBox::Ok);
  about.setInformativeText("<p align='center'>A Paraview Optimization GUI Plugin<br>Version: 0.2.2<br><br>© 2021 iDesign Lab, Iowa State University. All rights reserved.<br><br>Developers: Andrew Thomas and Ping He</p>");
  int ret = about.exec();
}

void pqOptAirfoil::caseFileHandler()
{
  paths casePaths = pvOptFuncs::linkCaseFile(this);
  *caseFileDir = casePaths.caseFileDir;
  rootPath = casePaths.rootPath;
  *pyGuiPath = pvOptFuncs::writePythonGUI(rootPath);
  dockerMessage();
}

void pqOptAirfoil::calculateAdvancedMeshSettings()
{
  // enable/disable inputs
  nTotalCells->setEnabled(!(*advancedSettingsEnabled));
  dX1PSValue->setEnabled(*advancedSettingsEnabled);
  dX2PSValue->setEnabled(*advancedSettingsEnabled);
  dXMaxPSValue->setEnabled(*advancedSettingsEnabled);
  dX1SSValue->setEnabled(*advancedSettingsEnabled);
  dX2SSValue->setEnabled(*advancedSettingsEnabled);
  dXMaxSSValue->setEnabled(*advancedSettingsEnabled);
  NpTEValue->setEnabled(*advancedSettingsEnabled);
  marchingDistanceValue->setEnabled(*advancedSettingsEnabled);

  if (!(*advancedSettingsEnabled))
  {
    double parameterScaleFactor = 125/(nTotalCells->text().toDouble()/NpExtrude->text().toDouble());
    // calculate presets for advanced settings
    dX1PSValue->setText(QString::fromStdString(std::to_string(0.005*parameterScaleFactor)));
    dX2PSValue->setText(QString::fromStdString(std::to_string(0.002*parameterScaleFactor)));
    dXMaxPSValue->setText(QString::fromStdString(std::to_string(0.02*parameterScaleFactor)));
    dX1SSValue->setText(QString::fromStdString(std::to_string(0.005*parameterScaleFactor)));
    dX2SSValue->setText(QString::fromStdString(std::to_string(0.002*parameterScaleFactor)));
    dXMaxSSValue->setText(QString::fromStdString(std::to_string(0.02*parameterScaleFactor)));
    NpTEValue->setText(QString::fromStdString(std::to_string(ceil(5/parameterScaleFactor))));
  }
  else
  {
    // calculate nTotalCells from advanced input
    calculateTotalCells();
  }
}

void pqOptAirfoil::calculateTotalCells()
{
  double psScalingFactor = 0.3333*((dX1PSValue->text().toDouble()/0.005)+(dX2PSValue->text().toDouble()/0.002)+(dXMaxPSValue->text().toDouble()/0.02));
  double ssScalingFactor = 0.3333*((dX1SSValue->text().toDouble()/0.005)+(dX2SSValue->text().toDouble()/0.002)+(dXMaxSSValue->text().toDouble()/0.02));
  nTotalCells->setText(std::to_string(floor(NpExtrude->text().toDouble()*125/(0.5*(ssScalingFactor+psScalingFactor)))).c_str());
}

void pqOptAirfoil::preProcessing()
{
  if (caseFileDir->size()<=1)
  {
    caseFileHandler();
  }
  if (upperSurfacePath->size()<=1)
  {
    selectUpperSurface();
  }
  if (lowerSurfacePath->size()<=1)
  {
    selectLowerSurface();
  }
  // write genAirfoilMesh.py
  writeGenMeshScript();
  //write other required files
  writeSystem();
  writeOrig();  

  std::ostringstream preProcessingPath;
  preProcessingPath << rootPath << preProcessingFile;

  #ifdef _WIN32
    std::ofstream preProcessing(preProcessingPath.str().c_str(), std::ios::binary);
  #else
    std::ofstream preProcessing(preProcessingPath.str().c_str(), std::ios::out);
  #endif
  
  //write preProcessing.sh
  preProcessing<<
  "#!/bin/bash\n"
  "\n"
  "\n"
  "if [ -z \"$WM_PROJECT\" ]; then\n"
  "  echo \"OpenFOAM environment not found\"\n"
  "  exit\n"
  "fi\n"
  "\n"
  "\n"
  "echo \"Generating mesh..\"\n"
  "python genAirfoilMesh.py &> "<<logMeshGenOut<<"\n"
  "plot3dToFoam -noBlank volumeMesh.xyz >> "<<logMeshGenOut<<"\n"
  "autoPatch 45 -overwrite >> "<<logMeshGenOut<<"\n"
  "createPatch -overwrite >> "<<logMeshGenOut<<"\n"
  "renumberMesh -overwrite >> "<<logMeshGenOut<<"\n"
  "echo \"Generating mesh.. Done!\"\n"
  "\n"
  "\n"
  "cp -r 0.orig 0" << std::endl;
  preProcessing.flush();
  preProcessing.close();

  *preProcessingComplete = true;
}

void pqOptAirfoil::genMesh()
{
  preProcessing();

  bool success = false;
  std::string detailedOutput;

  std::ostringstream preProcessingPath;
  preProcessingPath << rootPath << preProcessingFile;

  std::string chmodCmd = "chmod -R 755 " + preProcessingPath.str();
  QFuture<void> futurePermission = QtConcurrent::run([=]() {
    pvOptFuncs::executeSystemCmd(chmodCmd);
  });
  futurePermission.waitForFinished();
  std::string cmdPreProcessing = "docker run -i --rm -u dafoamuser --mount \"type=bind,src=" + rootPath +
  ",target=/home/dafoamuser/mount\" -w /home/dafoamuser/mount dafoam/opt-packages:v" + *daVersion + " bash -c \". /home/dafoamuser/dafoam/loadDAFoam.sh && ./preProcessing.sh\"";
  QFuture<void> futureMeshGen = QtConcurrent::run([=]() {
    pvOptFuncs::executeSystemCmd(cmdPreProcessing);
  });
  futureMeshGen.waitForFinished();
  //popup showing mesh generation output
  std::string logPath = rootPath + "/" + logMeshGenOut;
  success = true;
  QFile file(logPath.c_str());
  if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
  {
    success = false;
  }
  while (!file.atEnd()) 
  {
    QByteArray line = file.readLine();
    detailedOutput.append(line.toStdString().c_str());
  }
  file.close();

  QMessageBox meshGenInfo;
  meshGenInfo.setText(success ? tr("       Mesh Generation Complete       ") : (*goodDocker ? tr("       Error -- No Output File       ") : tr("Docker Not Installed")));
  meshGenInfo.setStandardButtons(QMessageBox::Ok);
  std::string infoMsg = "Created mesh with initial chord length: " + std::to_string(*originalChordValue) + "meters";
  if (*goodDocker)
  {
    meshGenInfo.setInformativeText(tr(infoMsg.c_str()));
    meshGenInfo.setDetailedText(QString(tr(detailedOutput.c_str())));
  }
  int ret = meshGenInfo.exec();

  if (success) 
  { 
    *meshGenComplete = true; 
  }
}

void pqOptAirfoil::selectUpperSurface()
{
  *upperSurfacePath = QFileDialog::getOpenFileName(this, tr("Select Upper Surface Profile"), QString(rootPath.c_str()));
  upperPath = upperSurfacePath->toStdString();
}

void pqOptAirfoil::selectLowerSurface()
{
  *lowerSurfacePath = QFileDialog::getOpenFileName(this, tr("Select Lower Surface Profile"), QString(rootPath.c_str()));
  lowerPath = lowerSurfacePath->toStdString();
}

void pqOptAirfoil::writeGenMeshScript()
{
  std::ostringstream genAirfoilMeshPath;
  genAirfoilMeshPath << rootPath << "/genAirfoilMesh.py";
  std::ofstream genAirfoilMeshScript(genAirfoilMeshPath.str().c_str(), std::ios::out);

  QString cutOff = "/";
  
  genAirfoilMeshScript<<
  "#!/usr/bin/env python\n"
  "\n"
  "\n"
  "from pyhyp import pyHyp\n"
  "import numpy\n"
  "from pyspline import *\n"
  "\n"
  "\n"
  "prefix = './profiles/'\n"
  "airfoilProfilePS = prefix + '"<<lowerSurfacePath->right(lowerSurfacePath->length() - lowerSurfacePath->lastIndexOf(cutOff) - 1).toStdString()<<"'\n"
  "airfoilProfileSS = prefix + '"<<upperSurfacePath->right(upperSurfacePath->length() - upperSurfacePath->lastIndexOf(cutOff) - 1).toStdString()<<"'\n"
  "ZSpan = 0.1\n"
  "nSpan = 2\n"
  "dX1PS = "<<dX1PSValue->text().toDouble()<<"\n"
  "Alpha1PS = 1.2\n"
  "dX2PS = "<<dX2PSValue->text().toDouble()<<"\n"
  "Alpha2PS = 1.2\n"
  "dXMaxPS = "<<dXMaxPSValue->text().toDouble()<<"\n"
  "\n"
  "dX1SS = "<<dX1SSValue->text().toDouble()<<"\n"
  "Alpha1SS = 1.2\n"
  "dX2SS = "<<dX2SSValue->text().toDouble()<<"\n"
  "Alpha2SS = 1.2\n"
  "dXMaxSS = "<<dXMaxSSValue->text().toDouble()<<"\n"
  "\n"
  "NpTE = "<<std::stoi(NpTEValue->text().toStdString())<<"\n"
  "\n"
  "NpExtrude = "<<std::stoi(NpExtrude->text().toStdString())<<"\n"
  "yWall = "<<yWall->text().toDouble()<<"\n"
  "marchDist = "<<marchingDistanceValue->text().toDouble()<<"\n"
  "\n"
  "\n"
  "\n"
  "fPS = open(airfoilProfilePS, 'r')\n"
  "linesPS = fPS.readlines()\n"
  "fPS.close()\n"
  "xPS = []\n"
  "yPS = []\n"
  "zPS = []\n"
  "for line in linesPS:\n"
  "    cols = line.split()\n"
  "    xPS.append(float(cols[0]))\n"
  "    yPS.append(float(cols[1]))\n"
  "\n"
  "for i in range(len(xPS)):\n"
  "    zPS.append(0.0)\n"
  "\n"
  "fSS = open(airfoilProfileSS, 'r')\n"
  "linesSS = fSS.readlines()\n"
  "fSS.close()\n"
  "xSS = []\n"
  "ySS = []\n"
  "zSS = []\n"
  "for line in linesSS:\n"
  "    cols = line.split()\n"
  "    xSS.append(float(cols[0]))\n"
  "    ySS.append(float(cols[1]))\n"
  "for i in range(len(xSS)):\n"
  "    zSS.append(0.0)\n"
  "\n"
  "\n"
  "\n"
  "tmp = dX1PS\n"
  "for i in range(1000):\n"
  "    if tmp > dXMaxPS:\n"
  "        nStretch1PS = i\n"
  "        break\n"
  "    else:\n"
  "        tmp = tmp * Alpha1PS\n"
  "tmp = dX2PS\n"
  "for i in range(1000):\n"
  "    if tmp > dXMaxPS:\n"
  "        nStretch2PS = i\n"
  "        break\n"
  "    else:\n"
  "        tmp = tmp * Alpha2PS\n"
  "\n"
  "xLPSConst = xPS[-1]\n"
  "xLPS1 = 0\n"
  "xLPS2 = 0\n"
  "for i in range(nStretch1PS):\n"
  "    xLPS1 += dX1PS * (Alpha1PS ** i)\n"
  "    xLPSConst -= dX1PS * (Alpha1PS ** i)\n"
  "for i in range(nStretch2PS):\n"
  "    xLPS2 += dX2PS * (Alpha2PS ** i)\n"
  "    xLPSConst -= dX2PS * (Alpha2PS ** i)\n"
  "nXConstPS = int(xLPSConst / dXMaxPS)\n"
  "dXMaxPS = xLPSConst / nXConstPS\n"
  "\n"
  "xInterpPS = [0]\n"
  "tmp = dX1PS\n"
  "for i in range(nStretch1PS):\n"
  "    xInterpPS.append(xInterpPS[-1] + tmp)\n"
  "    tmp = tmp * Alpha1PS\n"
  "for i in range(nXConstPS):\n"
  "    xInterpPS.append(xInterpPS[-1] + dXMaxPS)\n"
  "tmp = dX2PS * (Alpha2PS ** (nStretch2PS - 1))\n"
  "for i in range(nStretch2PS):\n"
  "    xInterpPS.append(xInterpPS[-1] + tmp)\n"
  "    tmp /= Alpha2PS\n"
  "c1PS = pySpline.Curve(x=xPS, y=yPS, z=zPS, k=3)\n"
  "XPS = c1PS(xInterpPS)\n"
  "c2PS = pySpline.Curve(X=XPS, k=3)\n"
  "x1PS = c2PS.X[:, 0]\n"
  "y1PS = c2PS.X[:, 1]\n"
  "\n"
  "\n"
  "\n"
  "tmp = dX1SS\n"
  "for i in range(1000):\n"
  "    if tmp > dXMaxSS:\n"
  "        nStretch1SS = i\n"
  "        break\n"
  "    else:\n"
  "        tmp = tmp * Alpha1SS\n"
  "tmp = dX2SS\n"
  "for i in range(1000):\n"
  "    if tmp > dXMaxSS:\n"
  "        nStretch2SS = i\n"
  "        break\n"
  "    else:\n"
  "        tmp = tmp * Alpha2SS\n"
  "\n"
  "xLSSConst = xSS[-1]\n"
  "xLSS1 = 0\n"
  "xLSS2 = 0\n"
  "for i in range(nStretch1SS):\n"
  "    xLSS1 += dX1SS * (Alpha1SS ** i)\n"
  "    xLSSConst -= dX1SS * (Alpha1SS ** i)\n"
  "for i in range(nStretch2SS):\n"
  "    xLSS2 += dX2SS * (Alpha2SS ** i)\n"
  "    xLSSConst -= dX2SS * (Alpha2SS ** i)\n"
  "nXConstSS = int(xLSSConst / dXMaxSS)\n"
  "dXMaxSS = xLSSConst / nXConstSS\n"
  "\n"
  "xInterpSS = [0]\n"
  "tmp = dX1SS\n"
  "for i in range(nStretch1SS):\n"
  "    xInterpSS.append(xInterpSS[-1] + tmp)\n"
  "    tmp = tmp * Alpha1SS\n"
  "for i in range(nXConstSS):\n"
  "    xInterpSS.append(xInterpSS[-1] + dXMaxSS)\n"
  "tmp = dX2SS * (Alpha2SS ** (nStretch2SS - 1))\n"
  "for i in range(nStretch2SS):\n"
  "    xInterpSS.append(xInterpSS[-1] + tmp)\n"
  "    tmp /= Alpha2SS\n"
  "c1SS = pySpline.Curve(x=xSS, y=ySS, z=zSS, k=3)\n"
  "XSS = c1SS(xInterpSS)\n"
  "c2SS = pySpline.Curve(X=XSS, k=3)\n"
  "x1SS = c2SS.X[:, 0]\n"
  "y1SS = c2SS.X[:, 1]\n"
  "\n"
  "delta_y = numpy.linspace(y1PS[-1], y1SS[-1], NpTE, 'd')\n"
  "delta_y = delta_y[1:]\n"
  "delta_x = numpy.ones_like(delta_y, 'd')\n"
  "for i in range(len(delta_x)):\n"
  "    delta_x[i] = x1SS[-1]\n"
  "\n"
  "x1SS_Flip = x1SS[::-1]\n"
  "xAll = numpy.append(x1SS_Flip, x1PS[1:])\n"
  "xAll = numpy.append(xAll, delta_x)\n"
  "\n"
  "y1SS_Flip = y1SS[::-1]\n"
  "yAll = numpy.append(y1SS_Flip, y1PS[1:])\n"
  "yAll = numpy.append(yAll, delta_y)\n"
  "\n"
  "\n"
  "print('nPoints for PS: ', nStretch1PS + nStretch2PS + nXConstPS)\n"
  "print('nPoints for SS: ', nStretch1SS + nStretch2SS + nXConstSS)\n"
  "print('nPoints for TE: ', NpTE)\n"
  "print('nPoints Total: ', nStretch1PS + nStretch2PS + nXConstPS + nStretch1SS + nStretch2SS + nXConstSS + NpTE)\n"
  "print(\n"
  "    'Mesh cells: ',\n"
  "    (nStretch1PS + nStretch2PS + nXConstPS + nStretch1SS + nStretch2SS + nXConstSS + NpTE - 1)\n"
  "    * (NpExtrude - 1)\n"
  "    * (nSpan - 1),\n"
  ")\n"
  "\n"
  "\n"
  "f = open('surfaceMesh.xyz', 'w')\n"
  "f.write('1\\n')\n"
  "f.write('%d %d %d\\n' % (len(xAll), nSpan, 1))\n"
  "for iDim in range(3):\n"
  "    for z in numpy.linspace(0.0, ZSpan, nSpan):\n"
  "        for i in range(len(xAll)):\n"
  "            if iDim == 0:\n"
  "                f.write('%20.16f\\n' % xAll[i])\n"
  "            elif iDim == 1:\n"
  "                f.write('%20.16f\\n' % yAll[i])\n"
  "            else:\n"
  "                f.write('%20.16f\\n' % z)\n"
  "f.close()\n"
  "\n"
  "\n"
  "options = {\n"
  "    #        Input Parameters\n"
  "    'inputFile': 'surfaceMesh.xyz',\n"
  "    'unattachedEdgesAreSymmetry': False,\n"
  "    'outerFaceBC': 'farField',\n"
  "    'autoConnect': True,\n"
  "    'BC': {1: {'jLow': 'zSymm', 'jHigh': 'zSymm'}},\n"
  "    'families': 'wall',\n"
  "    #        Grid Parameters\n"
  "    'N': NpExtrude,\n"
  "    's0': yWall,\n"
  "    'marchDist': marchDist,\n"
  "    #   Pseudo Grid Parameters\n"
  "    'ps0': -1,\n"
  "    'pGridRatio': -1,\n"
  "    'cMax': 0.5,\n"
  "    #   Smoothing parameters\n"
  "    'epsE': 2.0,\n"
  "    'epsI': 4.0,\n"
  "    'theta': 2.0,\n"
  "    'volCoef': 0.20,\n"
  "    'volBlend': 0.0005,\n"
  "    'volSmoothIter': 20,\n"
  "}\n"
  "\n"
  "hyp = pyHyp(options=options)\n"
  "hyp.run()\n"
  "# hyp.writeCGNS('volumeMesh.cgns')\n"
  "hyp.writePlot3D('volumeMesh.xyz')" << std::endl;
  genAirfoilMeshScript.flush();
  genAirfoilMeshScript.close();
}

void pqOptAirfoil::machWarning()
{
  if (machNumberValue->text() != tr(""))
  {
    std::string outMsg = "";
    double machNum = machNumberValue->text().toDouble();
    int solver = solverBox->currentIndex();
    if (machNum<=0)
    {
      outMsg = "Mach Number must be greater than zero";
    }
    else if (machNum>0)
    {
      if (machNum < 0.1 && solver != 0)
      {
        outMsg = "The recommended solver for mach number of 0-0.1 is DASimpleFoam.\n\n" + solverName + " is currently selected and may fail.";
      }
      else if (machNum < 0.6 && machNum >= 0.1 && solver != 1)
      {
        outMsg = "The recommended solver for mach number of 0.1-0.6 is DARhoSimpleFoam.\n\n"  + solverName + " is currently selected and may fail.";
      }
      else if (machNum < 0.999 && machNum >=0.6 && solver != 2)
      {
        outMsg = "The recommended solver for mach number of 0.6-0.99 is DARhoSimpleCFoam.\n\n" + solverName +  " is currently selected and may fail.";
      }
      else if (machNum >= 1.0)
      {
        outMsg = "Super-sonic may fail.";
      }
    }

    if (outMsg != "")
    {
      QMessageBox msgBox;
      msgBox.setText(tr("Warning"));
      msgBox.setInformativeText(tr(outMsg.c_str()));
      msgBox.setStandardButtons(QMessageBox::Ok);
      int ret = msgBox.exec();
    }
  }
}

void pqOptAirfoil::writeOrig()
{
  double T = temperatureValue->text().toDouble();
  double U = (solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? machNumberValue->text().toDouble()*343.0 : machNumberValue->text().toDouble()*sqrt(1.4*287.0*T));
  double p = farFieldPressureValue->text().toDouble();
  double r = eddyViscosityRatioValue->text().toDouble();
  double i = turbulentIntensityValue->text().toDouble();
  double k = 1.5*pow(U*i,2);
  double epsilon = 0.09*normalizationDensityValue->text().toDouble()*pow(k,2) / (r*dynamicViscosityValue->text().toDouble());
  double nut = r*dynamicViscosityValue->text().toDouble();
  double omega = k/(r*dynamicViscosityValue->text().toDouble());
  double alphaT = nut*(*prtValue);

  if (*processingComplete)
  {
    std::string origPath = rootPath + "/0.orig";
    bool removedDirectory = vtksys::SystemTools::RemoveADirectory(origPath.c_str());

    std::string oPath = rootPath + "/0";
    bool createdDirectory = vtksys::SystemTools::MakeDirectory(oPath.c_str());
  }

  std::string origPath = rootPath + "/0.orig";
  bool createdDirectory = vtksys::SystemTools::MakeDirectory(origPath.c_str());

  std::stringstream epsilonLine;
  epsilonLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       volScalarField;\n"
  "    location    \"0\";\n"
  "    object      epsilon;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "dimensions      [0 2 -3 0 0 0 0];\n"
  "\n"
  "internalField uniform "<<epsilon<<";\n"
  "\n"
  "boundaryField\n"
  "{\n"
  "    \"(wing"<<(solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? ".*" : "|wingte")<<")\"\n"
  "    {\n"
  "        type            epsilonWallFunction;\n"
  "        value           $internalField;\n"
  "    }\n"
  "    symmetry1\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    symmetry2\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    inout\n"
  "    {\n"
  "        type            inletOutlet;\n"
  "        inletValue      $internalField;\n"
  "        value           $internalField;\n"
  "    }\n"
  "}\n"
  "\n"
  "// ************************************************************************* //";

  std::ostringstream epsilonPath;
  epsilonPath << origPath << "/epsilon";
  std::ofstream epsilonFile(epsilonLine.str().c_str(), std::ios::out);
  epsilonFile<<epsilonLine.str()<<std::endl;
  epsilonFile.flush();
  epsilonFile.close();


  std::stringstream kLine;
  kLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       volScalarField;\n"
  "    location    \"0\";\n"
  "    object      k;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "dimensions      [0 2 -2 0 0 0 0];\n"
  "\n"
  "internalField   uniform "<<k<<";\n"
  "\n"
  "boundaryField\n"
  "{\n"
  "    \"(wing.*)\"\n"
  "    {\n"
  "        type            kqRWallFunction;\n"
  "        value           $internalField;\n"
  "    }\n"
  "    symmetry1\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    symmetry2\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    inout\n"
  "    {\n"
  "        type            inletOutlet;\n"
  "        inletValue      $internalField;\n"
  "        value           $internalField;\n"
  "    }\n"
  "}\n"
  "\n"
  "\n"
  "\n"
  "// ************************************************************************* //";
  std::ostringstream kPath;
  kPath << origPath << "/k";
  std::ofstream kFile(kPath.str().c_str(), std::ios::out);
  kFile<<kLine.str()<<std::endl;
  kFile.flush();
  kFile.close();

  
  std::stringstream nutLine;
  nutLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       volScalarField;\n"
  "    location    \"0\";\n"
  "    object      nut;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "dimensions      [0 2 -1 0 0 0 0];\n"
  "\n"
  "internalField   uniform "<<nut<<";\n"
  "\n"
  "boundaryField\n"
  "{\n"
  "    \"(wing.*)\"\n"
  "    {\n"
  "        type            nutUSpaldingWallFunction;\n"
  "        value           $internalField;\n"
  "    }\n"
  "    symmetry1\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    symmetry2\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    inout\n"
  "    {\n"
  "        type            calculated;\n"
  "        value           $internalField;\n"
  "    }\n"
  "\n"
  "}\n"
  "\n"
  "\n"
  "// ************************************************************************* //";
  std::ostringstream nutPath;
  nutPath << origPath << "/nut";
  std::ofstream nutFile(nutPath.str().c_str(), std::ios::out);
  nutFile<<nutLine.str()<<std::endl;
  nutFile.flush();
  nutFile.close();
  
  
  std::stringstream nuTildaLine;
  nuTildaLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       volScalarField;\n"
  "    location    \"0\";\n"
  "    object      nuTilda;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "dimensions      [0 2 -1 0 0 0 0];\n"
  "\n"
  "internalField   uniform "<<nut<<";\n"
  "\n"
  "boundaryField\n"
  "{\n"
  "    \"(wing.*)\"\n"
  "    {\n"
  "        type            fixedValue;\n"
  "        value           uniform 0.0;\n"
  "    }\n"
  "    symmetry1\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    symmetry2\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    inout\n"
  "    {\n"
  "        type            inletOutlet;\n"
  "        inletValue      $internalField;\n"
  "        value           $internalField;\n"
  "    }\n"
  "}\n"
  "\n"
  "\n"
  "// ************************************************************************* //";
  std::ostringstream nuTildaPath;
  nuTildaPath << origPath << "/nuTilda";
  std::ofstream nuTildaFile(nuTildaPath.str().c_str(), std::ios::out);
  nuTildaFile<<nuTildaLine.str()<<std::endl;
  nuTildaFile.flush();
  nuTildaFile.close();


  std::stringstream omegaLine;
  omegaLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       volScalarField;\n"
  "    location    \"0\";\n"
  "    object      omega;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "dimensions      [0 0 -1 0 0 0 0];\n"
  "\n"
  "internalField   uniform "<<omega<<";\n"
  "\n"
  "boundaryField\n"
  "{\n"
  "    \"(wing"<<(solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? ".*" : "|wingte")<<")\"\n"
  "    {\n"
  "        type            omegaWallFunction;\n"
  "        value           $internalField;\n"
  "        blended         true;\n"
  "    }\n"
  "    symmetry1\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    symmetry2\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    inout\n"
  "    {\n"
  "        type            inletOutlet;\n"
  "        inletValue      $internalField;\n"
  "        value           $internalField;\n"
  "    }\n"
  "}\n"
  "\n"
  "\n"
  "\n"
  "// ************************************************************************* //";
  std::ostringstream omegaPath;
  omegaPath << origPath << "/omega";
  std::ofstream omegaFile(omegaPath.str().c_str(), std::ios::out);
  omegaFile<<omegaLine.str()<<std::endl;
  omegaFile.flush();
  omegaFile.close();
  

  std::stringstream uLine;
  uLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       volVectorField;\n"
  "    location    \"0\";\n"
  "    object      U;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "dimensions      [0 1 -1 0 0 0 0];\n"
  "\n"
  "internalField uniform ("<<U<<" 0 0);\n"
  "\n"
  "boundaryField\n"
  "{\n"
  "    \"(wing.*)\"\n"
  "    {\n"
  "        type            fixedValue;\n"
  "        value           uniform (0 0 0);\n"
  "    }\n"
  "    symmetry1\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    symmetry2\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    inout\n"
  "    {\n"
  "        type            inletOutlet;\n"
  "        inletValue      $internalField;\n"
  "        value           $internalField;\n"
  "    }\n"
  "}\n"
  "\n"
  "\n"
  "// ************************************************************************* //";
  std::ostringstream uPath;
  uPath << origPath << "/U";
  std::ofstream uFile(uPath.str().c_str(), std::ios::out);
  uFile<<uLine.str()<<std::endl;
  uFile.flush();
  uFile.close();


  std::stringstream pLine;
  pLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       volScalarField;\n"
  "    location    \"0\";\n"
  "    object      p;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "dimensions      "<<(solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? "[0 2 -2 0 0 0 0]" : "[1 -1 -2 0 0 0 0]")<<";\n"
  "\n"
  "internalField   uniform "<<p<<";\n"
  "\n"
  "boundaryField\n"
  "{\n"
  "    \"(wing.*)\"\n"
  "    {\n"
  "        type            zeroGradient;\n"
  "    }\n"
  "    symmetry1\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    symmetry2\n"
  "    {\n"
  "        type            symmetry;\n"
  "    }\n"
  "    inout\n"
  "    {\n"
  "        type            fixedValue;\n"
  "        value           $internalField;\n"
  "    }\n"
  "}\n"
  "\n"
  "\n"
  "// ************************************************************************* //";
  std::ostringstream pPath;
  pPath << origPath << "/p";
  std::ofstream pFile(pPath.str().c_str(), std::ios::out);
  pFile<<pLine.str()<<std::endl;
  pFile.flush();
  pFile.close();

  if(solverBox->currentText() != tr("DASimpleFoam (incompressible)"))
  {
    std::stringstream alphatLine;
    alphatLine<<
    "FoamFile\n"
    "{\n"
    "    version     2.0;\n"
    "    format      ascii;\n"
    "    class       volScalarField;\n"
    "    location    \"0\";\n"
    "    object      alphat;\n"
    "}\n"
    "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    "\n"
    "dimensions      [1 -1 -1 0 0 0 0];\n"
    "\n"
    "internalField   uniform "<<alphaT<<";\n"
    "\n"
    "boundaryField\n"
    "{\n"
    "    \"wing.*\"\n"
    "    {\n"
    "        type            compressible::alphatWallFunction;\n"
    "        value           $internalField;\n"
    "    }\n"
    "    symmetry1\n"
    "    {\n"
    "        type            symmetry;\n"
    "    }\n"
    "    symmetry2\n"
    "    {\n"
    "        type            symmetry;\n"
    "    }\n"
    "    inout\n"
    "    {\n"
    "        type            calculated;\n"
    "        value           $internalField;\n"
    "    }\n"
    "}\n"
    "\n"
    "\n"
    "// ************************************************************************* //";
    std::ostringstream alphatPath;
    alphatPath << origPath << "/alphat";
    std::ofstream alphatFile(alphatPath.str().c_str(), std::ios::out);
    alphatFile<<alphatLine.str()<<std::endl;
    alphatFile.flush();
    alphatFile.close();

    
    std::stringstream tLine;
    tLine<<
    "FoamFile\n"
    "{\n"
    "    version     2.0;\n"
    "    format      ascii;\n"
    "    class       volScalarField;\n"
    "    location    \"0\";\n"
    "    object      T;\n"
    "}\n"
    "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    "\n"
    "dimensions      [0 0 0 1 0 0 0];\n"
    "\n"
    "internalField   uniform "<<T<<";\n"
    "\n"
    "boundaryField\n"
    "{\n"
    "    \"wing.*\"\n"
    "    {\n"
    "        type            fixedValue;\n"
    "        value           $internalField;\n"
    "    }\n"
    "    symmetry1\n"
    "    {\n"
    "        type            symmetry;\n"
    "    }\n"
    "    symmetry2\n"
    "    {\n"
    "        type            symmetry;\n"
    "    }\n"
    "    inout\n"
    "    {\n"
    "        type            inletOutlet;\n"
    "        inletValue      $internalField;\n"
    "        value           $internalField;\n"
    "    }\n"
    "}\n"
    "\n"
    "\n"
    "// ************************************************************************* //";
    std::ostringstream tPath;
    tPath << origPath << "/T";
    std::ofstream tFile(tPath.str().c_str(), std::ios::out);
    tFile<<tLine.str()<<std::endl;
    tFile.flush();
    tFile.close();

    if (*processingComplete)
    {
      std::ostringstream alphatPath;
      alphatPath << rootPath << "/0/alphat";
      std::ofstream alphatFile(alphatPath.str().c_str(), std::ios::out);
      alphatFile<<alphatLine.str()<<std::endl;
      alphatFile.flush();
      alphatFile.close();

      std::ostringstream tPath;
      tPath << rootPath << "/0/T";
      std::ofstream tFile(tPath.str().c_str(), std::ios::out);
      tFile<<tLine.str()<<std::endl;
      tFile.flush();
      tFile.close();
    }
  }


  if (*processingComplete)
  {
    std::ostringstream epsilonPath;
    epsilonPath << rootPath << "/0/epsilon";
    std::ofstream epsilonFile(epsilonPath.str().c_str(), std::ios::out);
    epsilonFile<<epsilonLine.str()<<std::endl;
    epsilonFile.flush();
    epsilonFile.close();

    std::ostringstream kPath;
    kPath << rootPath << "/0/k";
    std::ofstream kFile(kPath.str().c_str(), std::ios::out);
    kFile<<kLine.str()<<std::endl;
    kFile.flush();
    kFile.close();
    
    std::ostringstream nutPath;
    nutPath << rootPath << "/0/nut";
    std::ofstream nutFile(nutPath.str().c_str(), std::ios::out);
    nutFile<<nutLine.str()<<std::endl;
    nutFile.flush();
    nutFile.close();

    std::ostringstream nuTildaPath;
    nuTildaPath << rootPath << "/0/nuTilda";
    std::ofstream nuTildaFile(nuTildaPath.str().c_str(), std::ios::out);
    nuTildaFile<<nuTildaLine.str()<<std::endl;
    nuTildaFile.flush();
    nuTildaFile.close();

    std::ostringstream omegaPath;
    omegaPath << rootPath << "/0/omega";
    std::ofstream omegaFile(omegaPath.str().c_str(), std::ios::out);
    omegaFile<<omegaLine.str()<<std::endl;
    omegaFile.flush();
    omegaFile.close();

    std::ostringstream uPath;
    uPath << rootPath << "/0/U";
    std::ofstream uFile(uPath.str().c_str(), std::ios::out);
    uFile<<uLine.str()<<std::endl;
    uFile.flush();
    uFile.close();
    
    std::ostringstream pPath;
    pPath << rootPath << "/0/p";
    std::ofstream pFile(pPath.str().c_str(), std::ios::out);
    pFile<<pLine.str()<<std::endl;
    pFile.flush();
    pFile.close();
  }
}

void pqOptAirfoil::writeConstant()
{
  std::string constantPath = rootPath + "/constant";
  bool createdDirectory = vtksys::SystemTools::MakeDirectory(constantPath.c_str());

  std::stringstream transportLine;
  transportLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"constant\";\n"
  "    object      transportProperties;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "transportModel Newtonian;\n"
  "\n"
  "nu "<<dynamicViscosityValue->text().toDouble()<<";\n"
  "Pr 0.7;\n"
  "Prt "<<*prtValue<<";\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //";

  std::stringstream thermoLine;
  thermoLine<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"constant\";\n"
  "    object      thermophysicalProperties;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "thermoType\n"
  "{\n"
  "    mixture               pureMixture;\n"
  "    specie                specie;\n"
  "    equationOfState       perfectGas;\n"
  "    energy                sensibleInternalEnergy;\n"
  "    thermo                hConst;\n"
  "    type                  hePsiThermo;\n"
  "    transport             const;\n"
  "}\n"
  "\n"
  "mixture\n"
  "{\n"
  "    specie\n"
  "    {\n"
  "        molWeight           28.970000; \n"
  "    }\n"
  "    thermodynamics\n"
  "    {\n"
  "        Cp                  1005.000000;\n"
  "        Hf                  0.000000; \n"
  "    } \n"
  "    transport \n"
  "    { \n"
  "        mu                  "<<dynamicViscosityValue->text().toDouble()*normalizationDensityValue->text().toDouble()<<"; \n" //nu * rho
  "        Pr                  0.700000;\n"
  "        TRef                "<<temperatureValue->text().toDouble()<<";\n"
  "    }\n"
  "}\n"
  "\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //";
  std::ostringstream transportPath;
  transportPath << rootPath << "/constant" << (solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? "/transportProperties" : "/thermophysicalProperties");
  std::ofstream transportFile(transportPath.str().c_str(), std::ios::out);
  transportFile<<(solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? transportLine.str() : thermoLine.str())<<std::endl;
  transportFile.flush();
  transportFile.close();

  std::stringstream prtVal;
  prtVal<<"    Prt                  "<<(*prtValue)<<";\n";

  std::ostringstream turbulencePath;
  turbulencePath << rootPath << "/constant/turbulenceProperties";
  std::ofstream turbulenceFile(turbulencePath.str().c_str(), std::ios::out);
  turbulenceFile<<
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"constant\";\n"
  "    object      turbulenceProperties;\n"
  "}\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
  "\n"
  "simulationType RAS;\n"
  "RAS \n"
  "{ \n"
  "    RASModel             "<<(turbulenceModelBox->currentText() == tr("Spalart Allmaras") ? "SpalartAllmaras" : (turbulenceModelBox->currentText() == tr("k epsilon") ? "kEpsilon" : "kOmegaSST"))<<";\n"
  "    turbulence           on;\n"
  "    printCoeffs          off;\n"<<
  (solverBox->currentText() != tr("DASimpleFoam (incompressible)") ? prtVal.str() : "")<<"} \n"
  "\n"
  "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //" << std::endl;
  turbulenceFile.flush();
  turbulenceFile.close();
}


void pqOptAirfoil::writeSystem()
{
  std::string systemPath = rootPath + "/system";
  bool createdDirectory = vtksys::SystemTools::MakeDirectory(systemPath.c_str());

  std::ostringstream controlDictPath;
  controlDictPath << systemPath << "/controlDict";
  std::ofstream controlDictFile(controlDictPath.str().c_str(), std::ios::out);

  controlDictFile << 
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"system\";\n"
  "    object      controlDict;\n"
  "}\n"
  "application     simpleDAFoam;\n"
  "startFrom       startTime;\n"
  "startTime       0;\n"
  "stopAt          endTime;\n"
  "endTime         "<<std::stoi(flowIterationsValue->text().toStdString())<<";\n"
  "deltaT          1;\n"
  "writeControl    timeStep;\n"
  "writeInterval   "<<std::stoi(flowIterationsValue->text().toStdString())<<";\n"
  "purgeWrite      0;\n"
  "writeFormat     ascii;\n"
  "writePrecision  8;\n"
  "writeCompression on;\n"
  "timeFormat      general;\n"
  "timePrecision   8;\n"
  "runTimeModifiable true;\n"
  "DebugSwitches\n"
  "{\n"
  "    SolverPerformance 0;\n"
  "}\n";
  controlDictFile.flush();
  controlDictFile.close();


  std::ostringstream createPatchDictPath;
  createPatchDictPath << systemPath << "/createPatchDict";
  std::ofstream createPatchDictFile(createPatchDictPath.str().c_str(), std::ios::out);
  createPatchDictFile << 
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    object      createPatchDict;\n"
  "}\n"
  "pointSync false;\n"
  "patches\n"
  "(\n"
  "    {\n"
  "        name symmetry1;\n"
  "        patchInfo\n"
  "        {\n"
  "            type symmetry;\n"
  "        }\n"
  "        constructFrom patches;\n"
  "        patches (auto0);\n"
  "    }\n"
  "    {\n"
  "        name symmetry2;\n"
  "        patchInfo\n"
  "        {\n"
  "            type symmetry;\n"
  "        }\n"
  "        constructFrom patches;\n"
  "        patches (auto1);\n"
  "    }\n"
  "\n"
  "    {\n"
  "        name wing;\n"
  "        patchInfo\n"
  "        {\n"
  "            type wall;\n"
  "        }\n"
  "        constructFrom patches;\n"
  "        patches (auto2 auto3);\n"
  "    }\n"
  "\n"
  "    {\n"
  "        name inout;\n"
  "        patchInfo\n"
  "        {\n"
  "            type patch;\n"
  "        }\n"
  "        constructFrom patches;\n"
  "        patches (auto4);\n"
  "    }\n"
  ");\n";
  createPatchDictFile.flush();
  createPatchDictFile.close();

  std::ostringstream fvSchemesPath;
  fvSchemesPath << systemPath << "/fvSchemes";
  std::ofstream fvSchemesFile(fvSchemesPath.str().c_str(), std::ios::out);
  fvSchemesFile << 
    "FoamFile\n"
    "{\n"
    "    version     2.0;\n"
    "    format      ascii;\n"
    "    class       dictionary;\n"
    "    object      fvSchemes;\n"
    "}\n"
    "ddtSchemes \n"
    "{\n"
    "    default                                             steadyState;\n"
    "}\n"
    "gradSchemes\n"
    "{\n"
    "    default                                             Gauss linear;\n"
    "}\n"
    "interpolationSchemes\n"
    "{\n"
    "    default                                             linear;\n"
    "}\n"
    "laplacianSchemes\n"
    "{\n"
    "    default                                             Gauss linear corrected;\n"
    "}\n"
    "snGradSchemes\n"
    "{\n"
    "    default                                             corrected;\n"
    "}\n"
    "wallDist\n"
    "{\n"
    "    method                                              meshWave;\n"
    "}\n";

  if (solverBox->currentText() != tr("DARhoSimpleCFoam (transonic)"))
  {
    fvSchemesFile << 
    "divSchemes\n"
    "{\n"
    "    default                                             none;\n"
    "    div(phi,U)                                          bounded Gauss linearUpwindV grad(U);\n"
    "    div(phi,e)                                          bounded Gauss upwind;\n"
    "    div(phi,h)                                          bounded Gauss upwind;\n"
    "    div(phi,T)                                          bounded Gauss upwind;\n"
    "    div(phi,nuTilda)                                    bounded Gauss upwind;\n"
    "    div(phi,k)                                          bounded Gauss upwind;\n"
    "    div(phi,omega)                                      bounded Gauss upwind;\n"
    "    div(phi,epsilon)                                    bounded Gauss upwind;\n"
    "    div(phi,K)                                          bounded Gauss upwind;\n"
    "    div(phi,Ekp)                                        bounded Gauss upwind;\n"
    "    div((nuEff*dev2(T(grad(U)))))                       Gauss linear;\n"
    "    div(((rho*nuEff)*dev2(T(grad(U)))))                 Gauss linear;\n"
    "    div(pc)                                             bounded Gauss upwind;\n"
    "}\n";
  }
  else
  {
    fvSchemesFile << 
    "divSchemes\n"
    "    {\n"
    "    default                                             none;\n"
    "    div(phi,U)                                          Gauss linearUpwindV grad(U);\n"
    "    div(phi,e)                                          Gauss upwind;\n"
    "    div((nuEff*dev2(T(grad(U)))))                       Gauss linear;\n"
    "    div(phi,h)                                          Gauss upwind;\n"
    "    div(phid,p)                                         Gauss limitedLinear 1.0;\n"
    "    div(((rho*nuEff)*dev2(T(grad(U)))))                 Gauss linear;\n"
    "    div(phi,nuTilda)                                    Gauss upwind;\n"
    "    div(phi,k)                                          Gauss upwind;\n"
    "    div(phi,omega)                                      Gauss upwind;\n"
    "    div(phi,epsilon)                                    Gauss upwind;\n"
    "    div(phi,K)                                          Gauss upwind;\n"
    "    div(phi,Ekp)                                        Gauss upwind;\n"
    "    div(pc)                                             Gauss upwind;\n"
    "}\n";
  }
  fvSchemesFile.flush();
  fvSchemesFile.close();

  std::ostringstream fvSolutionPath;
  fvSolutionPath << systemPath << "/fvSolution";
  std::ofstream fvSolutionFile(fvSolutionPath.str().c_str(), std::ios::out);
  fvSolutionFile <<
    "FoamFile\n"
    "{\n"
    "    version     2.0;\n"
    "    format      ascii;\n"
    "    class       dictionary;\n"
    "    location    \"system\";\n"
    "    object      fvSolution;\n"
    "}\n"
    "SIMPLE\n"
    "{\n"
    "    nNonOrthogonalCorrectors           0;\n"
    "}\n"
    "solvers\n"
    "{\n"
    "    \"(p|p_rgh|G)\"\n"
    "    {\n"
    "        \n"
    "        solver                         GAMG;\n"
    "        smoother                       GaussSeidel;\n"
    "        relTol                         0.1;\n"
    "        tolerance                      0;\n"
    "    }\n"
    "    Phi\n"
    "    {\n"
    "        $p;\n"
    "        relTol                         0;\n"
    "        tolerance                      1e-6;\n"
    "    }\n"
    "    \"(U|T|e|h|nuTilda|k|omega|epsilon)\"\n"
    "    {\n"
    "        solver                         smoothSolver;\n"
    "        smoother                       GaussSeidel;\n"
    "        relTol                         0.1;\n"
    "        tolerance                      0;\n"
    "        nSweeps                        1;\n"
    "    }\n"
    "}\n"
    "potentialFlow\n"
    "{\n"
    "    nNonOrthogonalCorrectors           20;\n"
    "}  \n";
  if (solverBox->currentText() != tr("DARhoSimpleCFoam (transonic)"))
  {
    fvSolutionFile <<
      "relaxationFactors\n"
      "{\n"
      "    fields\n"
      "    {\n"
      "        \"(p|p_rgh|rho)\"                     0.30;\n"
      "    }\n"
      "    equations\n"
      "    {\n"
      "        \"(U|T|e|h|nuTilda|k|epsilon|omega)\" 0.70;\n"
      "    }\n"
      "\n"
      "}\n" << std::endl;
  }
  else
  {
    fvSolutionFile <<
      "relaxationFactors\n"
      "{\n"
      "    fields\n"
      "    {\n"
      "        \"(p|p_rgh|rho)\"                     1.0;\n"
      "    }\n"
      "    equations\n"
      "    {\n"
      "        p                                     1.0;\n"
      "        \"(U|T|e|h|nuTilda|k|epsilon|omega)\" 0.80;\n"
      "    }\n"
      "\n"
      "}\n" << std::endl;
  }

  fvSolutionFile.flush();
  fvSolutionFile.close();
}

void pqOptAirfoil::toggleAdvancedTurbulence()
{
  if (*advancedTurbulenceEnabled)
  {
    *advancedTurbulenceEnabled = false;
  }
  else
  {
    *advancedTurbulenceEnabled = true;
  }
  caseSpecificParameters();
}

// case specific parameters
void pqOptAirfoil::caseSpecificParameters()
{
  // solver specific
  temperatureValue->setEnabled(solverBox->currentText() != tr("DASimpleFoam (incompressible)"));
  farFieldPressureValue->setEnabled(solverBox->currentText() != tr("DASimpleFoam (incompressible)"));
  farFieldPressureValue->setText((solverBox->currentText() != tr("DASimpleFoam (incompressible)")) ? (farFieldPressureValue->text() == "0.0" ? "101325" : farFieldPressureValue->text()) : "0.0");
  normalizationDensityValue->setEnabled((solverBox->currentText() != tr("DASimpleFoam (incompressible)")));
  
  // turbulence model specific
  turbulentIntensityValue->setEnabled(turbulenceModelBox->currentText() != tr("Spalart Allmaras") && (*advancedTurbulenceEnabled));
  eddyViscosityRatioValue->setEnabled((*advancedTurbulenceEnabled));

  if (solverBox->currentText() == tr("DASimpleFoam (incompressible)"))
  {
    solverName = "DASimpleFoam";
  }
  else if (solverBox->currentText() == tr("DARhoSimpleFoam (subsonic)"))
  {
    solverName = "DARhoSimpleFoam";
  }
  else
  {
    solverName = "DARhoSimpleCFoam";
  }

}

void pqOptAirfoil::toggleAdvancedRunscript()
{
  if (*advancedRunscriptEnabled)
  {
    *advancedRunscriptEnabled = false;
  }
  else
  {
    *advancedRunscriptEnabled = true;
  }
  advancedRunscript();
}

void pqOptAirfoil::advancedRunscript()
{
  dynamicViscosityValue->setEnabled((*advancedRunscriptEnabled));
  flowConvergenceToleranceValue->setEnabled((*advancedRunscriptEnabled));
  adjointConvergenceToleranceValue->setEnabled((*advancedRunscriptEnabled));
  feasibilityTolerance->setEnabled((*advancedRunscriptEnabled));
  optimalityTolerance->setEnabled((*advancedRunscriptEnabled));
  minimumThicknessValue->setEnabled((*advancedRunscriptEnabled));
  maximumThicknessValue->setEnabled((*advancedRunscriptEnabled));
  minimumVolumeValue->setEnabled((*advancedRunscriptEnabled));
  maximumVolumeValue->setEnabled((*advancedRunscriptEnabled));
  customDaConfigValue->setEnabled((*advancedRunscriptEnabled));
}

void pqOptAirfoil::scaleMesh()
{
  if (reynoldsNumberValue->text() != tr(""))
  {
    double temperature = temperatureValue->text().toDouble();
    double velocity = (solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? machNumberValue->text().toDouble()*343.0 : machNumberValue->text().toDouble()*sqrt(1.4*287.0*(temperature)));
    double dynamicViscosity = dynamicViscosityValue->text().toDouble();

    if (*goodDocker && *meshGenComplete)
    {
      if (caseFileDir->count()<=1)
      {
        caseFileHandler();
      }
      if (!(*profileReferenced))
      {
        getReferenceChord();
      }

      QMessageBox meshScalePrompt;
      meshScalePrompt.setText(tr("       Scale Parameter?       "));
      QPushButton* viscosityButton = meshScalePrompt.addButton(tr("Viscosity"), QMessageBox::ActionRole);
      QPushButton* scaleMeshButton = meshScalePrompt.addButton(tr("Geometry"), QMessageBox::ActionRole);

      int ret = meshScalePrompt.exec();

      if (*chordValue > 0 && velocity != 0 && normalizationDensityValue->text().toDouble() > 0 && reynoldsNumberValue->text().toDouble() > 0 && meshScalePrompt.clickedButton() == viscosityButton)
      {
        double newViscosity = normalizationDensityValue->text().toDouble() * velocity * (*chordValue) / reynoldsNumberValue->text().toDouble();
        dynamicViscosityValue->setText(QString(std::to_string(newViscosity).c_str()));
        QMessageBox viscScaleInfo;
        std::string info = "      Viscosity Scaled to " + std::to_string(newViscosity) + "      ";
        viscScaleInfo.setText(tr(info.c_str()));
        int ret = viscScaleInfo.exec();
      }

      if (*chordValue > 0 && velocity != 0 && normalizationDensityValue->text().toDouble() > 0 && reynoldsNumberValue->text().toDouble() > 0 && meshScalePrompt.clickedButton() == scaleMeshButton)
      { 
        std::string tempFile = "scaleHist.txt";
        *scaleFactor = ((reynoldsNumberValue->text().toDouble())*(dynamicViscosity))/((normalizationDensityValue->text().toDouble())*velocity*(*chordValue));
        std::string scaleCmd = "docker run -i --rm -u dafoamuser --mount \"type=bind,src=" + rootPath + 
        ",target=/home/dafoamuser/mount\" -w /home/dafoamuser/mount dafoam/opt-packages:v" + *daVersion + " bash -c \". /home/dafoamuser/dafoam/loadDAFoam.sh && transformPoints -scale '" + 
        std::to_string(*scaleFactor) + "' > " + tempFile.c_str() + "\"";
        QFuture<void> future = QtConcurrent::run([=]() {
          pvOptFuncs::executeSystemCmd(scaleCmd);
        });
        future.waitForFinished();
        *chordValue = *chordValue * *scaleFactor;

         if (*processingComplete)
        {
          scaleFFDPoints();
        }

        // popup showing mesh scaling output
        std::string logPath = rootPath + "/" + tempFile;
        std::string detailedOutput;
        bool success = true;
        QFile file(logPath.c_str());

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        {
          success = false;
        }
        while (!file.atEnd()) 
        {
          QByteArray line = file.readLine();
          detailedOutput.append(line.toStdString().c_str());
        }
        file.close();
        file.remove();

         QMessageBox meshScaleInfo;
        meshScaleInfo.setText(success ? tr("       Mesh Scaling Complete       ") : tr("       Error Opening Temporary File       "));
        std::string infoMsg = "Scaled mesh uniformly by: " + std::to_string(*scaleFactor);
        meshScaleInfo.setInformativeText(tr(infoMsg.c_str()));
        meshScaleInfo.setStandardButtons(QMessageBox::Ok);
        meshScaleInfo.setDetailedText(QString(tr(detailedOutput.c_str())));
        int ret = meshScaleInfo.exec();
      }
    }
    else
    {
      double newViscosity = normalizationDensityValue->text().toDouble() * velocity * (*chordValue) / reynoldsNumberValue->text().toDouble();
      dynamicViscosityValue->setText(QString(std::to_string(newViscosity).c_str()));
      QMessageBox viscScale;
      std::string info = "      Viscosity Scaled to " + std::to_string(newViscosity) + "      ";
      viscScale.setText(tr(info.c_str()));
      int ret = viscScale.exec();
    }
  }
}

void pqOptAirfoil::scaleReynolds()
{
  double velocity = (solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? machNumberValue->text().toDouble()*343.0 : machNumberValue->text().toDouble()*sqrt(1.4*287.0*temperatureValue->text().toDouble()));
  double newReynolds = normalizationDensityValue->text().toDouble() * velocity * (*chordValue) / dynamicViscosityValue->text().toDouble();
  reynoldsNumberValue->setText(QString(std::to_string(newReynolds).c_str()));
}

void pqOptAirfoil::getReferenceChord()
{
  *profileReferenced = true; 
  std::string profilePath;
  if (lowerSurfacePath->size()>1)
  {
    profilePath = lowerPath;
  }
  else if (upperSurfacePath->size()>1)
  {
    profilePath = upperPath;
  }
  else
  {
    selectLowerSurface();
    profilePath = lowerPath;
  }

  std::ifstream profile;
  profile.open (profilePath.c_str());

  std::string line;
  double largestXCoord = 0;
  double smallestXCoord = 0;
  while (std::getline(profile, line))
  {
    std::istringstream iss(line);
    std::vector<std::string> results(std::istream_iterator<std::string>{iss},
                               std::istream_iterator<std::string>());
    double xCoord = ::atof(results[0].c_str());
    largestXCoord = (xCoord > largestXCoord ? xCoord : largestXCoord);
    smallestXCoord = (xCoord < smallestXCoord ? xCoord : smallestXCoord);
  }
  *originalChordValue = largestXCoord - smallestXCoord;
  *chordValue = largestXCoord - smallestXCoord;
  profile.close();
}

void pqOptAirfoil::writeRunScript()
{
  if (caseFileDir->count()<=1)
  {
    caseFileHandler();
  }
  if (!(*profileReferenced))
  {
    getReferenceChord();
  }
  if (!(*preProcessingComplete))
  {
    preProcessing();
  }

  std::ostringstream runScriptPath;
  runScriptPath << rootPath << "/runScript.py";
  std::ofstream runScript(runScriptPath.str().c_str(), std::ios::out);

  double uRef = solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? machNumberValue->text().toDouble()*343.0 : machNumberValue->text().toDouble()*sqrt(1.4*287.0*(temperatureValue->text().toDouble()));
  double i = turbulentIntensityValue->text().toDouble();
  double r = eddyViscosityRatioValue->text().toDouble();
  double k = 1.5*pow(uRef*i,2);
  double edgeWidth = 1e-3;
  double sF = ((*chordValue)/(*originalChordValue));
  *referenceAreaValue = 0.1*(*chordValue);
  std::string customDaConfigString = "    " + customDaConfigValue->text().toStdString() + ",\n";

  // write run script
  runScript<<
  "#!/usr/bin/env python\n"
  "\n"
  "\n"
  "import os\n"
  "import argparse\n"
  "from mpi4py import MPI\n"
  "from dafoam import PYDAFOAM, optFuncs\n"
  "from pygeo import *\n"
  "from pyspline import *\n"
  "from idwarp import *\n"
  "from pyoptsparse import Optimization, OPT\n"
  "import numpy as np\n"
  "\n"
  "\n"
  "\n"
  "parser = argparse.ArgumentParser()\n"
  "parser.add_argument('--opt', help='optimizer to use', type=str, default='"<<optimizerBox->currentText().toStdString()<<"')\n"
  "parser.add_argument('--task', help='type of run to do', type=str, default='"<<(runType->currentText()==tr("Optimize") ? "opt" : "runPrimal")<<"')\n"
  "args = parser.parse_args()\n"
  "gcomm = MPI.COMM_WORLD\n"
  "\n"
  "U0 = "<<uRef<<"\n"
  "p0 = "<<farFieldPressureValue->text().toDouble()<<"\n"
  "nuTilda0 = "<<r*dynamicViscosityValue->text().toDouble()<<"\n"
  "k0 = "<<k<<"\n"
  "epsilon0 = "<<(0.09)*(normalizationDensityValue->text().toDouble()*pow(k,2)/(dynamicViscosityValue->text().toDouble()*r))<<"\n"
  "omega0 = "<<k/(dynamicViscosityValue->text().toDouble()*r)<<"\n"
  "T0 = "<<temperatureValue->text().toDouble()<<"\n"
  "CL_target = "<<targetLiftCoefficientValue->text().toDouble()<<"\n"
  "alpha0 = "<<alphaValue->text().toDouble()<<"\n"
  "rho0 = "<<normalizationDensityValue->text().toDouble()<<"\n"
  "A0 = "<<*referenceAreaValue<<"\n"
  "\n"
  "\n"
  "\n"
  "daOptions = {\n"
  "    'designSurfaces': ['wing'],\n"
  "    'solverName': '"<<solverName<<"',\n"
  "    'primalMinResTol': "<<flowConvergenceToleranceValue->text().toDouble()<<",\n"
  "    'printInterval' : "<<*printIntervalValue<<",\n"
  "    'primalBC': {\n"
  "        'U0': {'variable': 'U', 'patches': ['inout'], 'value': [U0, 0.0, 0.0]},\n"
  "        'p0': {'variable': 'p', 'patches': ['inout'], 'value': [p0]},\n"
  "        'T0': {'variable': 'T', 'patches': ['inout'], 'value': [T0]},\n"
  "        'nuTilda0': {'variable': 'nuTilda', 'patches': ['inout'], 'value': [nuTilda0]},\n"
  "        'k0': {'variable': 'k', 'patches': ['inout'], 'value': [k0]},\n"
  "        'omega0': {'variable': 'omega', 'patches': ['inout'], 'value': [omega0]},\n"
  "        'epsilon0': {'variable': 'epsilon', 'patches': ['inout'], 'value': [epsilon0]},\n"
  "        'useWallFunction': True,\n"
  "    },\n"<<(solverBox->currentText() != tr("DASimpleFoam (incompressible)") ? "    # variable bounds for compressible flow conditions\n"
  "    'primalVarBounds': {\n"
  "        'UMax': 1000.0,\n"
  "        'UMin': -1000.0,\n"
  "        'pMax': 500000.0,\n"
  "        'pMin': 20000.0,\n"
  "        'eMax': 500000.0,\n"
  "        'eMin': 100000.0,\n"
  "        'rhoMax': 5.0,\n"
  "        'rhoMin': 0.2}," : "\n")<<
  "    'objFunc': {\n"
  "        'CD': {\n"
  "            'part1': {\n"
  "                'type': 'force',\n"
  "                'source': 'patchToFace',\n"
  "                'patches': ['wing'],\n"
  "                'directionMode': 'parallelToFlow',\n"
  "                'alphaName': 'alpha',\n"
  "                'scale': 1.0 / (0.5 "<<(solverBox->currentText() != tr("DASimpleFoam (incompressible)") ? "* rho0 " : "")<<"* U0 * U0 * A0),\n"
  "                'addToAdjoint': True,\n"
  "            }\n"
  "        },\n"
  "        'CL': {\n"
  "            'part1': {\n"
  "                'type': 'force',\n"
  "                'source': 'patchToFace',\n"
  "                'patches': ['wing'],\n"
  "                'directionMode': 'normalToFlow',\n"
  "                'alphaName': 'alpha',\n"
  "                'scale': 1.0 / (0.5 "<<(solverBox->currentText() != tr("DASimpleFoam (incompressible)") ? "* rho0 " : "")<<"* U0 * U0 * A0),\n"
  "                'addToAdjoint': True,\n"
  "            }\n"
  "        },\n"
  "    },\n"
  "    'adjEqnOption': {'gmresRelTol': "<<adjointConvergenceToleranceValue->text().toDouble()<<", 'pcFillLevel': 1, 'jacMatReOrdering': 'rcm'},\n"<<
  (solverBox->currentText() == tr("DARhoSimpleCFoam (transonic)") ? 
  "    'transonicPCOption': 1," : "\n")<<
  "    'normalizeStates': {\n"
  "        'U': U0,\n"
  "        'p': "<<(solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? "U0 * U0 / 2.0" : "p0")<<",\n"
  "        'nuTilda': nuTilda0 * 10.0,\n"
  "        'k': k0 "<<(solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? "* 10.0" : "")<<",\n"
  "        'epsilon': epsilon0 "<<(solverBox->currentText() == tr("DASimpleFoam (incompressible)") ? "" : (solverBox->currentText() == tr("DARhoSimpleFoam (subsonic)") ? "* 0.1" : "* 0.01"))<<",\n"
  "        'omega': omega0 "<<(solverBox->currentText() != tr("DASimpleFoam (incompressible)") ? "* 0.01" : "")<<",\n"
  "        'phi': 1.0,\n"
  "    },\n"
  "    'checkMeshThreshold': {'maxAspectRatio': 100000.0}, \n"
  "    'adjPartDerivFDStep': {'State': 1e-6, 'FFD': 1e-3},\n"
  "    'primalMinResTolDiff': 1.0e5, \n"
  "    'adjPCLag': 100,\n"
  "    'designVar': {},\n"<<
  (customDaConfigValue->text().toStdString() != "" ? customDaConfigString : "")<<
  "}\n"
  "\n"
  "\n"
  "meshOptions = {\n"
  "    'gridFile': os.getcwd(),\n"
  "    'fileType': 'openfoam',\n"
  "    'symmetryPlanes': [[[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], [[0.0, 0.0, 0.1], [0.0, 0.0, 1.0]]],\n"
  "}\n"
  "\n"
  "\n"
  "if args.opt == 'snopt':\n"
  "    optOptions = {\n"
  "        'Major feasibility tolerance': "<<feasibilityTolerance->text().toDouble()<<",\n"
  "        'Major optimality tolerance': "<<optimalityTolerance->text().toDouble()<<",\n"
  "        'Function precision': 1.0e-7,\n"
  "        'Function precision': 1.0e-7,\n"
  "        'Verify level': -1,\n"
  "        'Major iterations limit': "<<std::stoi(maximumIterationsValue->text().toStdString())<<",\n"
  "        'Nonderivative linesearch': None,\n"
  "        'Print file': 'opt_SNOPT_print.out',\n"
  "        'Summary file': 'opt_SNOPT_summary.out',\n"
  "    }\n"
  "elif args.opt == 'ipopt':\n"
  "    optOptions = {\n"
  "       'tol': "<<optimalityTolerance->text().toDouble()<<",\n"
  "       'constr_viol_tol': "<<feasibilityTolerance->text().toDouble()<<",\n"
  "       'max_iter': "<<std::stoi(maximumIterationsValue->text().toStdString())<<",\n"
  "       'output_file': 'opt_IPOPT.txt',\n"
  "       'mu_strategy': 'adaptive',\n"
  "       'limited_memory_max_history': 10,\n"
  "       'nlp_scaling_method': 'none',\n"
  "       'alpha_for_y': 'full',\n"
  "       'recalc_y': 'yes',\n"
  "  }\n"
  "elif args.opt == 'slsqp':\n"
  "    optOptions = {\n"
  "        'ACC': 1.0e-7,\n"
  "        'MAXIT': "<<std::stoi(maximumIterationsValue->text().toStdString())<<",\n"
  "        'IFILE': 'opt_SLSQP.out',\n"
  "    }\n"
  "else:\n"
  "    print('opt arg not valid!')\n"
  "    exit(0)\n"
  "\n"
  "\n"
  "def alpha(val, geo):\n"
  "    aoa = val[0] * np.pi / 180.0\n"
  "    inletU = [float(U0 * np.cos(aoa)), float(U0 * np.sin(aoa)), 0]\n"
  "    DASolver.setOption('primalBC', {'U0': {'variable': 'U', 'patches': ['inout'], 'value': inletU}})\n"
  "    DASolver.updateDAOption()\n"
  "\n"
  "DVGeo = DVGeometry('./FFD/wingFFD.xyz')\n"
  "DVGeo.addRefAxis('bodyAxis', xFraction=0.25, alignIndex='k')\n"
  "iVol = 0\n"
  "pts = DVGeo.getLocalIndex(iVol)\n"
  "indexList = pts[:, :, :].flatten()\n"
  "PS = geo_utils.PointSelect('list', indexList)\n"
  "DVGeo.addGeoDVLocal('shapey', lower=-1.0, upper=1.0, axis='y', scale=1.0, pointSelect=PS)\n"
  "daOptions['designVar']['shapey'] = {'designVarType': 'FFD'}\n"
  "DVGeo.addGeoDVGlobal('alpha', [alpha0], alpha, lower=0.0, upper=10.0, scale=1.0)\n"
  "daOptions['designVar']['alpha'] = {'designVarType': 'AOA', 'patches': ['inout'], 'flowAxis': 'x', 'normalAxis': 'y'}\n"
  "\n"
  "\n"
  "\n"
  "DASolver = PYDAFOAM(options=daOptions, comm=gcomm)\n"
  "DASolver.setDVGeo(DVGeo)\n"
  "mesh = USMesh(options=meshOptions, comm=gcomm)\n"
  "DASolver.addFamilyGroup(DASolver.getOption('designSurfaceFamily'), DASolver.getOption('designSurfaces'))\n"
  "DASolver.printFamilyList()\n"
  "DASolver.setMesh(mesh)\n"
  "evalFuncs = []\n"
  "DASolver.setEvalFuncs(evalFuncs)\n"
  "\n"
  "\n"
  "\n"
  "DVCon = DVConstraints()\n"
  "DVCon.setDVGeo(DVGeo)\n"
  "DVCon.setSurface(DASolver.getTriangulatedMeshSurface(groupName=DASolver.getOption('designSurfaceFamily')))\n"
  "\n"
  "leList = [["<<edgeWidth*sF<<", "<<0.0<<", "<<edgeWidth*sF<<"], ["<<edgeWidth*sF<<", "<<0.0<<", "<<(0.1-edgeWidth)*sF<<"]]\n"
  "teList = [["<<(*originalChordValue - edgeWidth)*sF<<", "<<0.0<<", "<<edgeWidth*sF<<"], ["<<(*originalChordValue - edgeWidth)*sF<<", "<<0.0<<", "<<(0.1-edgeWidth)*sF<<"]]\n"
  "DVCon.addVolumeConstraint(leList, teList, nSpan=2, nChord=10, lower="<<minimumVolumeValue->text().toDouble()<<", upper="<<maximumVolumeValue->text().toDouble()<<", scaled=True)\n"
  "\n"
  "DVCon.addThicknessConstraints2D(leList, teList, nSpan=2, nChord=10, lower="<<minimumThicknessValue->text().toDouble()<<", upper="<<maximumThicknessValue->text().toDouble()<<", scaled=True)\n"
  "\n"
  "nFFDs_x = pts.shape[0]\n"
  "indSetA = []\n"
  "indSetB = []\n"
  "for i in range(nFFDs_x):\n"
  "    for j in [0, 1]:\n"
  "        indSetA.append(pts[i, j, 1])\n"
  "        indSetB.append(pts[i, j, 0])\n"
  "DVCon.addLinearConstraintsShape(indSetA, indSetB, factorA=1.0, factorB=-1.0, lower=0.0, upper=0.0)\n"
  "\n"
  "indSetA = []\n"
  "indSetB = []\n"
  "for i in [0, nFFDs_x - 1]:\n"
  "    for k in [0]:\n"
  "        indSetA.append(pts[i, 0, k])\n"
  "        indSetB.append(pts[i, 1, k])\n"
  "DVCon.addLinearConstraintsShape(indSetA, indSetB, factorA=1.0, factorB=1.0, lower=0.0, upper=0.0)\n"
  "\n"
  "\n"
  "optFuncs.DASolver = DASolver\n"
  "optFuncs.DVGeo = DVGeo\n"
  "optFuncs.DVCon = DVCon\n"
  "optFuncs.evalFuncs = evalFuncs\n"
  "optFuncs.gcomm = gcomm\n"
  "\n"
  "\n"
  "\n"
  "if args.task == 'opt':\n"
  "\n"
  "    alpha4CLTarget = optFuncs.solveCL(CL_target, 'alpha', 'CL')\n"
  "    alpha([alpha4CLTarget], None)\n"
  "\n"
  "    optProb = Optimization('opt', objFun=optFuncs.calcObjFuncValues, comm=gcomm)\n"
  "    DVGeo.addVariablesPyOpt(optProb)\n"
  "    DVCon.addConstraintsPyOpt(optProb)\n"
  "\n"
  "    optProb.addObj('CD', scale=1)\n"
  "    optProb.addCon('CL', lower=CL_target, upper=CL_target, scale=1)\n"
  "\n"
  "    if gcomm.rank == 0:\n"
  "        print(optProb)\n"
  "\n"
  "    DASolver.runColoring()\n"
  "\n"
  "    opt = OPT(args.opt, options=optOptions)\n"
  "    histFile = './%s_hist.hst' % args.opt\n"
  "    sol = opt(optProb, sens=optFuncs.calcObjFuncSens, storeHistory=histFile)\n"
  "    if gcomm.rank == 0:\n"
  "        print(sol)\n"
  "\n"
  "elif args.task == 'runPrimal':\n"
  "\n"
  "    optFuncs.runPrimal()\n"
  "\n"
  "else:\n"
  "    print('task arg not found!')\n"
  "    exit(0)" << std::endl;
  runScript.flush();
  runScript.close();

  genFFDPoints();
  if (*goodDocker)
  {
    scaleFFDPoints();
  }

  *processingComplete = true;

  // Rewrite necessary files with updated data
  std::string oPath = rootPath + "/0";
  bool removedDirectory = vtksys::SystemTools::RemoveADirectory(oPath.c_str());
  writeOrig();
  
  std::string transPath = rootPath + "/constant/transportProperties";
  if (solverBox->currentText() != tr("DASimpleFoam (incompressible)")) { bool removedFile = vtksys::SystemTools::RemoveFile(transPath.c_str()); }
  writeConstant();
  writeSystem();

  //popup showing mesh generation output
  std::string logPath = rootPath + "/runScript.py";
  std::string detailedOutput;
  bool success = true;
  QFile file(logPath.c_str());
  if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
  {
    success = false;
  }
  while (!file.atEnd()) 
  {
    QByteArray line = file.readLine();
    detailedOutput.append(line.toStdString().c_str());
  }
  file.flush();
  file.close();

  bool warning = (reynoldsNumberValue->text().toDouble() <= 0) || (machNumberValue->text().toDouble() <= 0);
  std::string warningMsg = "Warning, one or more required parameters are not set";

  QMessageBox runScriptInfo;
  runScriptInfo.setText(warning ? tr(warningMsg.c_str()) : (success ? tr("       Write Run Script Complete       ") : tr("       Error -- No Output File       ")));
  runScriptInfo.setStandardButtons(QMessageBox::Ok);
  runScriptInfo.setDetailedText(QString(tr(detailedOutput.c_str())));
  runScriptInfo.setInformativeText(*goodDocker ? tr("click 'Run Case' to run with one core") : tr("Docker not Found -- Run Optimization Disabled"));
  QPushButton* optimizeButton = runScriptInfo.addButton(tr("Run Case"), QMessageBox::ActionRole);
  int ret = runScriptInfo.exec();

  if (runScriptInfo.clickedButton() == optimizeButton && *goodDocker)
  {
    //2>&1 | tee -a
    std::string optCmd = "docker run -i --rm -u dafoamuser --mount \"type=bind,src=" + rootPath + 
    ",target=/home/dafoamuser/mount\" -w /home/dafoamuser/mount dafoam/opt-packages:v" + *daVersion + " bash -c \". /home/dafoamuser/dafoam/loadDAFoam.sh && mpirun -np 1 python runScript.py 2>&1 | tee -a " + *optOutputFile + "\"";
    QFuture<void> future = QtConcurrent::run([=]() {
      pvOptFuncs::executeSystemCmd(optCmd);
    });
    if (future.isRunning())
    {
      #ifdef _WIN32
        Sleep(int(5e2)); //millisec
      #else
        usleep(int(5e5)); //microsec
      #endif
      *optimizationRunning = true;
    }
  }
  if (*optimizationRunning) 
  {  
    postProcessHandler();  
  }
}

void pqOptAirfoil::genFFDPoints()
{
  std::vector<int> factors;
  int nz = 2;
  int ny = 2;
  int nx = 2;
  int nearestEven = 2*round(round((numberFFDPointsValue->text().toInt()/nz))/2);
  for (int i = 1; i<=nearestEven; ++i)
  {
    if (nearestEven % i == 0)
    {
      factors.push_back(i);
    }
  }
  int facLen = factors.size();
  int n;
  if (facLen > 1)
  {
    if (facLen <= 5)
    {
      n = round(floor(facLen/2 - 1));
      ny = factors.at(n);
      nx = factors.at(facLen-(n+1));
    }
    else
    {
      n = round(facLen/3);
      ny = factors.at(n);
      nx = factors.at(facLen-(n+1));
    }
  }
  numberFFDPointsValue->setText(QString(std::to_string(nx*ny*nz).c_str()));

  std::string ffdPath = rootPath + "/FFD";
  bool createdDirectory = vtksys::SystemTools::MakeDirectory(ffdPath.c_str());

  std::string genFFDFileName = "genFFD.py";
  std::ostringstream genFFDPath;
  genFFDPath << ffdPath << "/" << genFFDFileName;
  std::ofstream genFFD(genFFDPath.str().c_str(), std::ios::out);

  genFFD<<
  "import numpy as np\n"
  "import sys\n"
  "\n"
  "\n"
  "def writeFFDFile(fileName,nBlocks,nx,ny,nz,points):\n"
  "\n"
  "    f = open(fileName,'w')\n"
  "\n"
  "    f.write('%d\\n'%nBlocks)\n"
  "    for i in range(nBlocks):\n"
  "        f.write('%d %d %d '%(nx[i],ny[i],nz[i]))\n"
  "    # end\n"
  "    f.write('\\n')\n"
  "    for block in range(nBlocks):\n"
  "        for k in range(nz[block]):\n"
  "            for j in range(ny[block]):\n"
  "                for i in range(nx[block]):\n"
  "                    f.write('%f '%points[block][i,j,k,0])\n"
  "                # end\n"
  "            # end\n"
  "        # end\n"
  "        f.write('\\n')\n"
  "\n"
  "        for k in range(nz[block]):\n"
  "            for j in range(ny[block]):\n"
  "                for i in range(nx[block]):\n"
  "                    f.write('%f '%points[block][i,j,k,1])\n"
  "                # end\n"
  "            # end\n"
  "        # end\n"
  "        f.write('\\n')\n"
  "\n"
  "        for k in range(nz[block]):\n"
  "            for j in range(ny[block]):\n"
  "                for i in range(nx[block]):\n"
  "                    f.write('%f '%points[block][i,j,k,2])\n"
  "                # end\n"
  "            # end\n"
  "        # end\n"
  "    # end\n"
  "    f.close()\n"
  "    return\n"
  "\n"
  "def returnBlockPoints(corners,nx,ny,nz):\n"
  "\n"
  "    points = np.zeros([nx,ny,nz,3])\n"
  "\n"
  "    # points 1 - 4 are the iMin face\n"
  "    # points 5 - 8 are the iMax face\n"
  "\n"
  "    for idim in range(3):\n"
  "        edge1 = np.linspace(corners[0][idim],corners[4][idim],nx)\n"
  "        edge2 = np.linspace(corners[1][idim],corners[5][idim],nx)\n"
  "        edge3 = np.linspace(corners[2][idim],corners[6][idim],nx)\n"
  "        edge4 = np.linspace(corners[3][idim],corners[7][idim],nx)\n"
  "\n"
  "        for i in range(nx):\n"
  "            edge5 = np.linspace(edge1[i],edge3[i],ny)\n"
  "            edge6 = np.linspace(edge2[i],edge4[i],ny)\n"
  "            for j in range(ny):\n"
  "                edge7 = np.linspace(edge5[j],edge6[j],nz)\n"
  "                points[i,j,:,idim] = edge7\n"
  "            # end\n"
  "        # end\n"
  "    # end\n"
  "\n"
  "               \n"
  "    return points\n"
  "\n"
  "nBlocks = 1\n"
  "\n"
  "nx = ["<<nx<<"]\n"
  "ny = ["<<ny<<"]\n"
  "nz = ["<<nz<<"]\n"
  "\n"
  "corners = np.zeros([nBlocks,8,3])\n"
  "\n"
  "corners[0,0,:] = [-0.010,-.0700,0.0]\n"
  "corners[0,1,:] = [-0.010,-0.0700,0.1]\n"
  "corners[0,2,:] = [-0.010,0.0700,0.0]\n"
  "corners[0,3,:] = [-0.010,0.0700,0.1]\n"
  "corners[0,4,:] = [ 1.01,-.07,0.0]\n"
  "corners[0,5,:] = [ 1.01,-.07,0.10]\n"
  "corners[0,6,:] = [ 1.01,0.07,0.0]\n"
  "corners[0,7,:] = [ 1.01,0.07,0.1]\n"
  "\n"
  "\n"
  "points = []\n"
  "for block in range(nBlocks):\n"
  "    points.append(returnBlockPoints(corners[block],nx[block],ny[block],nz[block]))\n"
  "\n"
  "#print points\n"
  "fileName = 'wingFFD.xyz'\n"
  "writeFFDFile(fileName,nBlocks,nx,ny,nz,points)" << std::endl;
  genFFD.flush();
  genFFD.close();

  std::string genFFDCmd = "cd " + ffdPath + " && python " + genFFDFileName;
  QFuture<void> future = QtConcurrent::run([=]() {
    pvOptFuncs::executeSystemCmd(genFFDCmd);
    });
  future.waitForFinished();
}

void pqOptAirfoil::scaleFFDPoints()
{
  std::ostringstream ffdPath;
  ffdPath << rootPath << "/FFD/wingFFD.xyz";
  std::ifstream FFD(ffdPath.str().c_str(), std::ios::in);
  std::stringstream newFFDData;
  std::string line;
  int i = 0;
  while (std::getline(FFD, line))
  {
    std::stringstream iss(line);
    std::string newLine;
    if (i<2)
    {
      newFFDData<<line<<std::endl;
    }
    if (i>1)
    {
      for(std::string s;iss>>s;)
      {
        double newVal = std::atof(s.c_str())*((*chordValue)/(*originalChordValue));
        newLine += std::to_string(newVal) + " ";
      }
      newFFDData<<newLine<<std::endl;
    }
    i=i+1;
  }
  FFD.close();
  bool removedFile = vtksys::SystemTools::RemoveFile(ffdPath.str().c_str());
  std::ostringstream newFFDPath;
  newFFDPath << rootPath << "/FFD/wingFFD.xyz";
  std::ofstream newFFD(newFFDPath.str().c_str(), std::ios::out);
  newFFD<<newFFDData.str()<<std::endl;
  newFFD.flush();
  newFFD.close();
}

void pqOptAirfoil::postProcessHandler()
{
  if (caseFileDir->size() <= 1)
  {
    caseFileHandler();
  }
  //opt info
  *alphaTitle = "\"CL Solution\\nAlpha0: " + std::to_string(*alphaZeroValue) + 
                ", target CL: " + targetLiftCoefficientValue->text().toStdString() + "\"";
  *optTitle = "\"Optimization\\nMach Number: " + machNumberValue->text().toStdString() + 
              ", Reynolds Number: " + reynoldsNumberValue->text().toStdString() + 
              ", target CL: " + targetLiftCoefficientValue->text().toStdString() + 
              "\\nMesh Cells: " + nTotalCells->text().toStdString() + 
              ", Solver: " + solverName +
              ", Turbulence Model: " + turbulenceModelBox->currentText().toStdString() + "\"";
  *simTitle = "\"Solving Flow\\nAlpha: " + std::to_string(*alphaZeroValue) + 
              ", Mach Number : " + machNumberValue->text().toStdString() + 
              ", Reynolds Number: " + reynoldsNumberValue->text().toStdString() +
              "\\nMesh Cells: " + nTotalCells->text().toStdString() +
              ", Solver: " + solverName +
              ", Turbulence Model: " + turbulenceModelBox->currentText().toStdString() + "\"";
  //show external plot interface
  std::string interactivePlotCmd;
  std::string log = rootPath + "/" + *optOutputFile;
  if (*optimizationRunning)
  {
    interactivePlotCmd = "python " + *pyGuiPath + " --animated" + " -log " + log + " -clTitle " + *alphaTitle + " -simTitle " + *simTitle + " -optTitle " + *optTitle;
  }
  else
  {
    interactivePlotCmd = "python " + *pyGuiPath;
  }
  QFuture<void> future = QtConcurrent::run([=]() {
    pvOptFuncs::executeSystemCmd(interactivePlotCmd);
  });
}